from dtiplayground.dmri.common.tools.base import ExternalToolWrapper
from dtiplayground.dmri.common import measure_time 
import dtiplayground.dmri.common
from pathlib import Path 
import subprocess as sp
import os 

logger=dtiplayground.dmri.common.logger.write

class NIRALUtilities(ExternalToolWrapper):
    def __init__(self,binary_path= None, **kwargs): ## binary_path in this class is binary dir  
        self.binary_path=None
        if binary_path is not None:
            self.binary_path=binary_path
        elif 'softwares' in kwargs:
            self.binary_path=kwargs['softwares']['niral_utilities']['path']

        self.dev_mode=False

    def polydatatransform(self,arguments: list):
        binary_name='polydatatransform'
        self.setArguments(arguments)
        return self.execute(binary_name,arguments)    

    def polydatamerge(self,arguments: list):
        binary_name='polydatamerge'
        self.setArguments(arguments)
        return self.execute(binary_name,arguments)    

    @measure_time
    def execute(self,binary_name,arguments=None,stdin=None):
        binary=Path(self.binary_path).joinpath(binary_name).__str__()
        command=[binary]+self.getArguments()
        if arguments is not None: command=[binary]+arguments
        logger("{}".format(command))
        output=sp.run(command,capture_output=True,text=True,stdin=stdin)
        if self.dev_mode:
            logger("{}\n{} {}".format(output.args,output.stdout,output.stderr))
            output.check_returncode()
        return output  ## output.returncode, output.stdout output.stderr, output.args, output.check_returncode()

    @measure_time
    def execute_pipe(self,binary_name,stdin=None):
        binary=Path(self.binary_path).joinpath(binary_name).__str__()
        command=[binary]+self.getArguments()
        if arguments is not None: command=[binary]+arguments
        logger("{}".format(command))
        if stdin is None:
            pipe_output=sp.Popen(command,stdout=sp.PIPE)
        else:
            pipe_output=sp.Popen(command,stdin=stdin,stdout=sp.PIPE)
        return pipe_output 


from dtiplayground.dmri.common.tools.base import ExternalToolWrapper
from dtiplayground.dmri.common import measure_time 
import dtiplayground.dmri.common
from pathlib import Path 
import subprocess as sp
import os 

logger=dtiplayground.dmri.common.logger.write

class FSL(ExternalToolWrapper):
    def __init__(self,binary_path= None, **kwargs): ## binary_path in this class is binary dir e.g. $FSLHOME 
        self.binary_path=binary_path
        self.arguments=[]
        self.fslhome=binary_path
        # self.num_threads=4
        self.dev_mode=False
        os.environ['FSLDIR']=self.fslhome 
        os.environ['FSLOUTPUTTYPE']="NIFTI_GZ"
        # os.environ['OMP_NUM_THREADS']=str(self.num_threads)

    def _set_num_threads(self,nth):
        self.num_threads=nth 
        os.environ['OMP_NUM_THREADS']=str(self.num_threads)

    def fslmaths_ops(self,input_file,output_file,operation):
        binary_name='fslmaths'
        arguments=[
                input_file,
                "-T{}".format(operation),
                output_file
        ]
        self.setArguments(arguments)
        return self.execute(binary_name,arguments)    

    def fslmaths_threshold(self,input_file,output_file,threshold=0):
        binary_name='fslmaths'
        arguments=[
                input_file,
                "-thr",
                "{}".format(threshold),
                output_file
        ]
        self.setArguments(arguments)
        return self.execute(binary_name,arguments)    

    def eddy_quad(self,
                  input_base,  #input base is path+file name without extensions
                  idx, #volume index file
                  par,
                  mask,
                  bvals):
        binary_name='eddy_quad'
        arguments=[
                input_base,
                "-idx",idx,
                "-par",par,
                "-m",mask,
                "-b",bvals
        ]
        self.setArguments(arguments)
        return self.execute(binary_name,arguments)    


    def bet(self,
            inputfile,
            outputfile):
        binary_name='bet'
        arguments=[
                    inputfile,
                    outputfile,
                    '-m',
                    '-v'
                ]
        self.setArguments(arguments)
        return self.execute(binary_name,arguments)

    def fslmerge(self,
                 outputfilename,
                 inputfiles:list):
        binary_name='fslmerge'
        arguments=[
                '-t',outputfilename
        ] + inputfiles
        self.setArguments(arguments)
        return self.execute(binary_name,arguments)        

    def topup(self,
              imain,  # input image filename
              datain, # acqp params filename
              out,    # output basename (not a filename), path + basename(without extension)
              fout,   # field output filename (Hz)
              iout,   # movement corrected image output filename
              config):# config filename
        binary_name='topup'
        arguments=[
                '--imain={}'.format(imain),
                '--datain={}'.format(datain),
                '--out={}'.format(out),
                '--fout={}'.format(fout),
                '--iout={}'.format(iout),
                '--config={}'.format(config)
        ]
        self.setArguments(arguments)
        return self.execute(binary_name,arguments)

    def eddy_openmp(self,
                    imain, # input image filename
                    mask, # brain mask filename
                    acqp, #acqp filename , topup params for b0 
                    index_file, #b0 index filename
                    bvals, #bvals filename
                    bvecs, #bvecs filename
                    out, # output basename (not a filename), path + basename(without extension)
                    estimate_move_by_susceptibility:bool = False,  # susceptibility correction
                    topup=None, # topuped file (if susceptibility_correct==True)
                    data_is_shelled=True,
                    repol=True,
                    verbose=True
                    ):
        binary_name='eddy_openmp'
        arguments=[]
        if topup is not None:  # susceptibility correction 
            arguments=[
                '--imain={}'.format(imain),
                '--mask={}'.format(mask),
                '--acqp={}'.format(acqp),
                '--index={}'.format(index_file),
                '--bvals={}'.format(bvals),
                '--bvecs={}'.format(bvecs),
                '--out={}'.format(out),
                '--topup={}'.format(topup)
            ]
            if estimate_move_by_susceptibility:
                arguments.append('--estimate_move_by_susceptibility')
        else: ## singlefile eddy correction without susceptibility
            arguments=[
                '--imain={}'.format(imain),
                '--mask={}'.format(mask),
                '--acqp={}'.format(acqp),
                '--index={}'.format(index_file),
                '--bvals={}'.format(bvals),
                '--bvecs={}'.format(bvecs),
                '--out={}'.format(out)
            ]
        if data_is_shelled: arguments.append('--data_is_shelled')
        if verbose : arguments.append('--verbose')
        if repol: arguments.append('--repol')

        self.setArguments(arguments)
        return self.execute(binary_name,arguments)

    @measure_time
    def execute(self,binary_name,arguments=None,stdin=None):
        binary=Path(self.binary_path).joinpath('bin').joinpath(binary_name).__str__()
        command=[binary]+self.getArguments()
        if arguments is not None: command=[binary]+arguments
        logger("{}".format(command))
        output=sp.run(command,capture_output=True,text=True,stdin=stdin)
        if self.dev_mode:
            logger("{}\n{} {}".format(output.args,output.stdout,output.stderr))
            output.check_returncode()
        return output  ## output.returncode, output.stdout output.stderr, output.args, output.check_returncode()

    @measure_time
    def execute_pipe(self,binary_name,stdin=None):
        binary=Path(self.binary_path).joinpath('bin').joinpath(binary_name).__str__()
        command=[binary]+self.getArguments()
        if arguments is not None: command=[binary]+arguments
        logger("{}".format(command))
        if stdin is None:
            pipe_output=sp.Popen(command,stdout=sp.PIPE)
        else:
            pipe_output=sp.Popen(command,stdin=stdin,stdout=sp.PIPE)
        return pipe_output 



  
import dtiplayground.dmri.preprocessing as prep
import yaml, os, time
from pathlib import Path
from copy import deepcopy
import numpy as np
logger=prep.logger.write

### dipy
from dipy.reconst.csdeconv import auto_response
from dipy.reconst.shm import CsaOdfModel
from dipy.data import default_sphere
from dipy.direction import peaks_from_model
from dipy.core.gradients import gradient_table

from dipy.reconst.dti import TensorFit, from_lower_triangular, decompose_tensor
from dipy.data import default_sphere, small_sphere
from dipy.direction.peaks import *
from dipy.direction.peaks import _pam_from_attrs

from dipy.reconst.csdeconv import (ConstrainedSphericalDeconvModel,
                                   auto_response_ssst)
from dipy.direction import DeterministicMaximumDirectionGetter, ProbabilisticDirectionGetter
from dipy.reconst.shm import CsaOdfModel

from dipy.tracking.stopping_criterion import ThresholdStoppingCriterion
from dipy.tracking import utils
from dipy.io.vtk import save_vtk_streamlines, load_vtk_streamlines

from dipy.tracking.local_tracking import LocalTracking
from dipy.tracking.streamline import Streamlines

@prep.measure_time
def peaks_from_dti(tensorfit,sphere, relative_peak_threshold,
                     min_separation_angle, mask=None, return_odf=False,
                     return_sh=True, gfa_thr=0, normalize_peaks=False,
                     sh_order=8, sh_basis_type=None, npeaks=5, B=None,
                     invB=None, parallel=False, num_processes=None):

    if return_sh and (B is None or invB is None):
        B, invB = sh_to_sf_matrix(
            sphere, sh_order, sh_basis_type, return_inv=True)

    shape = tensorfit.shape
    if mask is None:
        mask = np.ones(shape, dtype='bool')
    else:
        if mask.shape != shape:
            raise ValueError("Mask is not the same shape as data.")

    gfa_array = np.zeros(shape)
    qa_array = np.zeros((shape + (npeaks,)))

    peak_dirs = np.zeros((shape + (npeaks, 3)))
    peak_values = np.zeros((shape + (npeaks,)))
    peak_indices = np.zeros((shape + (npeaks,)), dtype='int')
    peak_indices.fill(-1)

    if return_sh:
        n_shm_coeff = (sh_order + 2) * (sh_order + 1) // 2
        shm_coeff = np.zeros((shape + (n_shm_coeff,)))

    if return_odf:
        odf_array = np.zeros((shape + (len(sphere.vertices),)))

    global_max = -np.inf
    # odf_all = tensorfit.odf(sphere)
    for idx in ndindex(shape):
        if not mask[idx]:
            continue

        #odf = model.fit(data[idx]).odf(sphere)
        odf = tensorfit[idx].odf(sphere)
        # odf = odf_all[idx]
        if return_sh:
            shm_coeff[idx] = np.dot(odf, invB)

        if return_odf:
            odf_array[idx] = odf

        gfa_array[idx] = gfa(odf)
        if gfa_array[idx] < gfa_thr:
            global_max = max(global_max, odf.max())
            continue

        # Get peaks of odf
        direction, pk, ind = peak_directions(odf, sphere,
                                             relative_peak_threshold,
                                             min_separation_angle)

        # Calculate peak metrics
        if pk.shape[0] != 0:
            global_max = max(global_max, pk[0])

            n = min(npeaks, pk.shape[0])
            qa_array[idx][:n] = pk[:n] - odf.min()

            peak_dirs[idx][:n] = direction[:n]
            peak_indices[idx][:n] = ind[:n]
            peak_values[idx][:n] = pk[:n]

            if normalize_peaks:
                peak_values[idx][:n] /= pk[0]
                peak_dirs[idx] *= peak_values[idx][:, None]

    qa_array /= global_max

    return _pam_from_attrs(PeaksAndMetrics,
                           sphere,
                           peak_indices,
                           peak_values,
                           peak_dirs,
                           gfa_array,
                           qa_array,
                           shm_coeff if return_sh else None,
                           B if return_sh else None,
                           odf_array if return_odf else None)



@prep.measure_time
def compute(*args,**kwargs):
    
    localTractography(*args)
    # dMaxTractography(*args)
    #probTractography(*args)
    exit(0)


@prep.measure_time
def localTractography(tensorImage, dwi_image, dipyLabelmap, output_dir,**kwargs):

    output_peaks_file = Path(output_dir).joinpath('peaks.pkl').__str__()
    output_sl_file = Path(output_dir).joinpath('streamlines.vtk').__str__()

    img = tensorImage.images
    img = np.moveaxis(img,3,0)
    logger("Computing dti parameters ...",prep.Color.PROCESS)
    dtiparams = np.zeros(img.shape[1:]+(12,))
    for idx in np.ndindex(img.shape[1:]):
        x,y,z = idx
        xx, xy, xz, yy,yz, zz=img[:,x,y,z]
        tri = np.array([xx,xy,yy,xz,yz,zz])
        lt=from_lower_triangular(tri)
        eigvals, eigvecs = decompose_tensor(lt)
        dtiparams[x,y,z,0:3] = eigvals
        dtiparams[x,y,z,3:] = eigvecs.ravel()


    img_label = dipyLabelmap
    affine=tensorImage.getAffineMatrixForNifti()
    labels=img_label

    tf = TensorFit(None,dtiparams)
    logger("Computing dti peaks ...",prep.Color.PROCESS)
    peaks = peaks_from_dti(  tf, default_sphere,
                             relative_peak_threshold=.8,
                             min_separation_angle=75,
                             mask=labels)


    stopping_criterion = ThresholdStoppingCriterion(peaks.gfa, .25)
    seed_mask = (labels == 1)
    seeds = utils.seeds_from_mask(seed_mask, affine, density=[3,3,3])
    logger("Generating streamlines ...",prep.Color.PROCESS)
    streamlines_generator = LocalTracking(peaks, stopping_criterion, seeds,
                                          affine=affine, step_size=.5)
    streamlines = Streamlines(streamlines_generator)

    logger("Saving streamlines ...",prep.Color.PROCESS)
    save_vtk_streamlines(streamlines, output_sl_file, to_lps=False, binary=False)
    logger("Tractogram generation completed",prep.Color.OK)


@prep.measure_time
def dMaxTractography(tensorfit, dwi_image, dipyLabelmap, output_dir,**kwargs):

    output_peaks_file = Path(output_dir).joinpath('peaks.pkl').__str__()
    output_sl_file = Path(output_dir).joinpath('streamlines.vtk').__str__()

    data = dwi_image.images
    affine = dwi_image.getAffineMatrixForNifti()
    bvecs = [x['unit_gradient'] for x in dwi_image.getGradients()]
    bvals = [x['b_value'] for x in dwi_image.getGradients()]
    gtab = gradient_table(bvals,bvecs)

    labels=dipyLabelmap
    seed_mask = labels == (labels == 1) | (labels == 2)
    white_matter = (labels == 1) | (labels == 2)
    # seed_mask = labels >= 0
    # white_matter = seed_mask
    seeds = utils.seeds_from_mask(seed_mask, affine, density=[1,1,1])
    response, ratio = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)

    csd_model = ConstrainedSphericalDeconvModel(gtab, response, sh_order=6)
    csd_fit = csd_model.fit(data, mask=white_matter)

    csa_model = CsaOdfModel(gtab, sh_order=6)
    gfa = csa_model.fit(data, mask=white_matter).gfa
    stopping_criterion = ThresholdStoppingCriterion(gfa, .25)

    detmax_dg = DeterministicMaximumDirectionGetter.from_shcoeff(
    csd_fit.shm_coeff, max_angle=70., sphere=default_sphere)
    streamline_generator = LocalTracking(detmax_dg, stopping_criterion, seeds,
                                         affine, step_size=.5)
    streamlines = Streamlines(streamline_generator)
    logger("Saving streamlines ...",prep.Color.PROCESS)
    save_vtk_streamlines(streamlines, output_sl_file, to_lps=False, binary=False)
    logger("Tractogram generation completed",prep.Color.OK)

@prep.measure_time
def probTractography(tensorfit, dwi_image, dipyLabelmap, output_dir,**kwargs):

    output_peaks_file = Path(output_dir).joinpath('peaks.pkl').__str__()
    output_sl_file = Path(output_dir).joinpath('streamlines.vtk').__str__()

    data = dwi_image.images
    affine = dwi_image.getAffineMatrixForNifti()
    bvecs = [x['unit_gradient'] for x in dwi_image.getGradients()]
    bvals = [x['b_value'] for x in dwi_image.getGradients()]
    gtab = gradient_table(bvals,bvecs)

    labels=dipyLabelmap
    seed_mask = labels == 1
    white_matter = (labels == 1) | (labels == 2)
    seeds = utils.seeds_from_mask(seed_mask, affine, density=[1,1,1])
    response, ratio = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)

    csd_model = ConstrainedSphericalDeconvModel(gtab, response, sh_order=6)
    csd_fit = csd_model.fit(data, mask=white_matter)

    csa_model = CsaOdfModel(gtab, sh_order=6)
    gfa = csa_model.fit(data, mask=white_matter).gfa
    stopping_criterion = ThresholdStoppingCriterion(gfa, .25)

    fod = csd_fit.odf(small_sphere)
    pmf = fod.clip(min=0)
    prob_dg = ProbabilisticDirectionGetter.from_pmf(pmf, max_angle=30.,
                                                    sphere=small_sphere)
    streamline_generator = LocalTracking(prob_dg, stopping_criterion, seeds,
                                         affine, step_size=.5)
    streamlines = Streamlines(streamline_generator)
    logger("Saving streamlines ...",prep.Color.PROCESS)
    save_vtk_streamlines(streamlines, output_sl_file, to_lps=False, binary=False)
    logger("Tractogram generation completed",prep.Color.OK)


import dtiplayground.dmri.common as common

logger=common.logger
measure_time=common.measure_time
Color=common.Color
get_uuid=common.get_uuid
object_by_id=common.object_by_id
get_timestamp=common.get_timestamp
    


from pathlib import Path 
import yaml
import json
import re

def load_bids(dirpaths:list):
  studies=[]
  hiddenRegex = re.compile(r'^\..*')
  notHidden = lambda x : hiddenRegex.search(x) is None

  for dr in dirpaths:
    d=Path(dr)
    study={ 'study_id': str(d), 
            'root_dir': str(d),
            'subjects':[]
          }

    subdirs = [ x for x in d.iterdir() if x.is_dir() and notHidden(x.name)]
    for sd in subdirs:
      subject = { 'subject_id': str(sd.relative_to(d)),
                  'sessions':[]
      }
      sesdirs = [ x for x in sd.iterdir() if x.is_dir() and notHidden(x.name)]
      for ssd in sesdirs:
        session = { 
                    'session_id': str(ssd.relative_to(sd)),
                    'modalities': []
        }
        mdirs = [ x for x in ssd.iterdir() if x.is_dir() and notHidden(x.name)]
        for md in mdirs:
          files = [ x for x in md.iterdir() if not x.is_dir() and notHidden(x.name)]
          image_files = list(filter(lambda x: ('.nrrd' in str(x.name)) or ('.nii' in str(x.name)), files))
          meta_files = list(map(lambda x: x.parent.joinpath(x.name.split('.')[0]+'.json'), image_files))
          file_infos = list(zip(image_files,meta_files))
          modality = {
            'modality_type' : str(md.relative_to(ssd)),
            'runs': []
          }
          for fi in file_infos:
            fn,mfn = fi
            if not fn.exists(): continue
            modalityfiles = list(filter(lambda x: mfn.stem in str(x) , files))
            run={
              'run_id': str(fn.relative_to(md)).split('.')[0],
              'files': list(map(lambda x: str(x.name),modalityfiles)),
              'attributes': None
            }
            if mfn.exists():
              run['meta_file']= str(mfn.relative_to(md))
            modality['runs'].append(run)
          session['modalities'].append(modality)
        subject['sessions'].append(session)
      study['subjects'].append(subject)
    studies.append(study)
  return studies 



from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
from PyQt5.QtCore import * 
import os
from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal as Signal

from dtiplayground.ui.module_selector_tab import ModuleSelector
from dtiplayground.ui.modules_details_tab import ModuleDetails
from dtiplayground.ui.general_tab import General

class ProtocolTabCommunicate(QObject):
  
  call_module_added_to_protocol = Signal()
  call_unsaved_modif = Signal()
  call_order_modules_in_protocol_list = Signal()
  set_loading_protocol = Signal(bool)
  get_module_id = Signal()
  set_module_id = Signal(int)
  call_set_default_protocol = Signal(str)
  call_read_protocol = Signal(str)
  call_remove_module_from_dic_protocol = Signal (int)
  enable_eddymotion_param = Signal(bool)
  set_dic_protocol = Signal(dict)
  call_no_unsaved_modif = Signal()
  call_write_protocol_yml = Signal(str)
  module_added_to_protocol_loading = Signal(list)
  call_update_user_preferences_file = Signal(dict)

  def CallModuleAddedToProtocol(self):
    self.call_module_added_to_protocol.emit()

  def CallUnsavedModif(self):
    self.call_unsaved_modif.emit()

  def CallOrderModulesInProtocolList(self):
    self.call_order_modules_in_protocol_list.emit()

  def SetLoadingProtocol(self, loading_protocol):
    self.set_loading_protocol.emit(loading_protocol)

  def GetModuleId(self):
    self.get_module_id.emit()
  
  def SetModuleId(self, new_id):
    self.set_module_id.emit(new_id)

  def CallSetDefaultProtocol(self, module_name):
    self.call_set_default_protocol.emit(module_name)

  def CallReadProtocol(self, filename):
    self.call_read_protocol.emit(filename)

  def CallRemoveModuleFromDicProtocol(self, key):
    self.call_remove_module_from_dic_protocol.emit(key)

  def EnableEddyMotionParam(self, value):
    self.enable_eddymotion_param.emit(value)

  def SetDicProtocol(self, new_dic_protocol):
    self.set_dic_protocol.emit(new_dic_protocol)

  def CallNoUnsavedModif(self):
    self.call_no_unsaved_modif.emit()

  def CallWriteProtocolYML(self, filename):
    self.call_write_protocol_yml.emit(filename)

  def ModuleAddedToProtocolLoading(self, list_modules):
    self.module_added_to_protocol_loading.emit(list_modules)

  def CallUpdateUserPreferencesFile(self, new_preferences):
    self.call_update_user_preferences_file.emit(new_preferences)


class ProtocolTab(QWidget):
  communicate = ProtocolTabCommunicate()

  def __init__(self, protocol_template, protocol_yml, preferences_yml):
    QWidget.__init__(self)
    self.preferences_yml = preferences_yml
    self.protocol_template = protocol_template
    self.tab = QWidget()

    # Tabs
    self.subtabs = QTabWidget()
    self.selector = ModuleSelector(self.protocol_template)
    self.details = ModuleDetails(self.protocol_template, preferences_yml) 
    self.general = General(self.protocol_template, protocol_yml) 
    self.subtabs.addTab(self.selector.tab, "Select")
    self.subtabs.addTab(self.details.tab, "Module Details")
    self.subtabs.addTab(self.general.tab, "General")
    self.subtabs.tabBarClicked.connect(self.details.ModuleDetailsClicked)    

    self.details.communicate.set_modules_details_tab.connect(self.SetDetailsDisplayTab)
    self.ProtocolTab()
    self.selector.communicate.set_drop_mode_protocol_list_widget.connect(lambda: self.protocol_list_widget.setDragDropMode(QAbstractItemView.DropOnly))
    self.ExcludePopup()
    self.MergePopup()

  def ProtocolTab(self):
    
    ### Protocol file path
    protocol_list_label = QLabel("Protocol:")
    if not hasattr(self, 'protocol_filename'):
      self.protocol_filename = os.getcwd() + "/protocol.yml"
    self.protocol_path_label = QLabel(self.protocol_filename)
    self.protocol_path_label.setWordWrap(True)

    ### Protocol List    
    self.protocol_list_widget = QListWidget()
    self.protocol_list_widget.setMouseTracking(True)
    self.protocol_list_widget.setSelectionMode(QAbstractItemView.ExtendedSelection)
    self.protocol_list_widget.setDragDropMode(QAbstractItemView.InternalMove) # default mode
    self.protocol_list_widget.itemPressed.connect(self.InternalMoveMode)
    self.protocol_list_widget.itemClicked.connect(self.GetSelectedItemData)
    self.protocol_list_widget.model().rowsInserted.connect(self.communicate.CallModuleAddedToProtocol)
    self.protocol_list_widget.model().rowsMoved.connect(self.communicate.CallUnsavedModif)
    self.protocol_list_widget.model().rowsMoved.connect(self.communicate.CallOrderModulesInProtocolList)

    ### Buttons
    # Default protocol button
    self.default_protocol_button = QPushButton("Default Protocol")
    self.default_protocol_button.clicked.connect(self.communicate.GetModuleId)
    self.default_protocol_button.setStatusTip("Sets the default protocol") 
    # Load protocol button
    self.load_protocol_button = QPushButton("Load Protocol")
    self.load_protocol_button.clicked.connect(self.LoadProtocol)
    self.load_protocol_button.setStatusTip("Load protocol file to use")    
    # Remove button
    self.remove_button = QPushButton("Remove selected module(s)")
    self.remove_button.clicked.connect(self.RemoveModule)
    self.remove_button.setStatusTip("Remove selected module(s) from the protocol list (DEL)") 
    # Clear button
    self.clear_button = QPushButton("Clear protocol")
    self.clear_button.clicked.connect(self.ClearProtocol)
    self.clear_button.setStatusTip("Remove all modules from the protocol list")    
    # Save button
    self.save_button = QPushButton("Save protocol")
    self.save_button.clicked.connect(self.SaveProtocol)
    self.save_button.setStatusTip("Save protocol file")    
    # Save As button
    self.saveas_button = QPushButton("Save protocol as...")
    self.saveas_button.clicked.connect(self.SaveAs)
    self.saveas_button.setStatusTip("Save protocol file as...")    

    ### Layout
    layout_protocol_h1 = QHBoxLayout()
    layout_protocol_h1.addWidget(self.remove_button)
    layout_protocol_h1.addWidget(self.clear_button)

    layout_protocol_v1 = QVBoxLayout()
    layout_protocol_v1.addWidget(protocol_list_label)
    layout_protocol_v1.addWidget(self.protocol_path_label)
    layout_protocol_v1.addWidget(self.default_protocol_button)
    layout_protocol_v1.addWidget(self.load_protocol_button)
    layout_protocol_v1.addWidget(self.protocol_list_widget)
    layout_protocol_v1.addLayout(layout_protocol_h1)
    layout_protocol_v1.addWidget(self.save_button)
    layout_protocol_v1.addWidget(self.saveas_button)
        
    layout_protocol_v2 = QVBoxLayout()
    layout_protocol_v2.addWidget(self.subtabs)

    layout_protocol = QHBoxLayout()
    layout_protocol.addLayout(layout_protocol_v1)
    layout_protocol.addLayout(layout_protocol_v2)
    self.tab.setLayout(layout_protocol)

  def InternalMoveMode(self):
    self.protocol_list_widget.setDragDropMode(QAbstractItemView.InternalMove)

  def DefaultProtocol(self, module_id):
    self.ClearProtocol()
    self.details.details_stack.setCurrentIndex(0)

    self.communicate.SetLoadingProtocol(True)

    for module_iterator_1 in self.protocol_template["options"]["execution"]["pipeline"]["default_value"]:
      for module_iterator_2 in self.protocol_template["options"]["execution"]["pipeline"]["candidates"]:
        if module_iterator_2["value"] == module_iterator_1:
          new_module = QListWidgetItem()
          new_module.setText(module_iterator_2["caption"])
          new_module.setData(QtCore.Qt.UserRole, module_id)
          new_module.setToolTip(module_iterator_2["description"])
          self.protocol_list_widget.addItem(new_module)
          module_id += 1

          self.communicate.CallSetDefaultProtocol(module_iterator_2["value"])
    self.communicate.SetModuleId(module_id)
    self.communicate.CallModuleAddedToProtocol()
    self.communicate.SetLoadingProtocol(False)

            
  def LoadProtocol(self):
    protocol_filter = "Protocol file (*.yaml *.yml)"
    input_protocol = QFileDialog.getOpenFileName(
      parent = self,
      caption = "Select a protocol file",
      filter = protocol_filter
    )
    filename = input_protocol[0]
    if filename != "":
      self.protocol_filename = filename
      self.protocol_path_label.setText(self.protocol_filename)
      self.communicate.CallReadProtocol(self.protocol_filename)
      self.details.details_stack.setCurrentIndex(0)

  def RemoveModule(self):

    self.communicate.CallOrderModulesInProtocolList()

    items = self.protocol_list_widget.selectedItems()

    for i in items:
      removed_item_index = self.protocol_list_widget.indexFromItem(i).row()
      removed_item_key = i.data(QtCore.Qt.UserRole)
      self.communicate.CallRemoveModuleFromDicProtocol(removed_item_key)
      removed_item = self.protocol_list_widget.takeItem(removed_item_index)
      if removed_item.text() == "Exclude Gradients":
        self.selector.EnableModulesListWidget(True)
      if removed_item.text() == "Susceptibility correction":
        self.communicate.EnableEddyMotionParam(False)

    if self.protocol_list_widget.count() == 0:
      self.selector.EnableExcludeGradientsModule(True)
      self.selector.EnableMergeImagesModule(True)
      self.selector.EnableModulesListWidget(True) 
      self.details.details_stack.setCurrentIndex(0)

    if len(items) > 0:
      self.communicate.CallUnsavedModif()
    
  def ClearProtocol(self):    
    self.subtabs.setCurrentIndex(0)

    if self.protocol_list_widget.count() > 0:
      self.communicate.CallUnsavedModif()

    for ite in range(self.protocol_list_widget.count()):
      removed_item = self.protocol_list_widget.takeItem(0)
    self.selector.EnableModulesListWidget(True)  
    self.selector.EnableExcludeGradientsModule(True)
    self.selector.EnableMergeImagesModule(True)
    self.communicate.EnableEddyMotionParam(False)
    self.communicate.SetDicProtocol({})

  def SaveAs(self):
    file_filter = "Protocol file (*.yml *.yaml)"
    filename = QFileDialog.getSaveFileName(
      parent = self,
      caption = "Select a protocol file",
      directory = "protocol.yml",
      filter = file_filter
      )
    if filename[0] != "":
      self.protocol_filename = filename[0]
      self.protocol_path_label.setText(self.protocol_filename)
      self.SaveProtocol()
      print("protocol saved as ", self.protocol_filename)

  def SaveProtocol(self):
    if not hasattr(self, 'protocol_filename'):
      self.protocol_filename = os.getcwd() + "/protocol.yml"
    self.communicate.CallOrderModulesInProtocolList()
    self.communicate.CallWriteProtocolYML(self.protocol_filename)
    self.communicate.CallNoUnsavedModif()

  def ExcludePopup(self):
    self.exclude_popup = QMessageBox()
    self.exclude_popup.setText("The Exclude Gradients module must run separately.")
    self.exclude_popup.setWindowTitle("DMRIPrep message")
    self.notshowagain_exclude = QCheckBox("Do not show again.")
    self.notshowagain_exclude.stateChanged.connect(self.NotShowAgainExcludePopupStateChanged)
    self.exclude_popup.setCheckBox(self.notshowagain_exclude)
    
  def NotShowAgainExcludePopupStateChanged(self):
    if self.notshowagain_exclude.isChecked():
      self.preferences_yml["showExcludeGradientsPopup"] = False
    else:
      self.preferences_yml["showExcludeGradientsPopup"] = True
    self.communicate.CallUpdateUserPreferencesFile(self.preferences_yml)

  def MergePopup(self):
    self.merge_popup = QMessageBox()
    self.merge_popup.setText("The Merge Images module must run separately.")
    self.merge_popup.setWindowTitle("DMRIPrep message")
    self.notshowagain_merge = QCheckBox("Do not show again.")
    self.notshowagain_merge.stateChanged.connect(self.NotShowAgainMergePopupStateChanged)
    self.merge_popup.setCheckBox(self.notshowagain_merge)
    
  def NotShowAgainMergePopupStateChanged(self):
    if self.notshowagain_merge.isChecked():
      self.preferences_yml["showMergeImagesPopup"] = False
    else:
      self.preferences_yml["showMergeImagesPopup"] = True
    self.communicate.CallUpdateUserPreferencesFile(self.preferences_yml)

  def SetDetailsDisplayTab(self):
    if self.subtabs.currentIndex() != 1: # if the active tab is not "Module Details"
      self.subtabs.setCurrentIndex(1)

  def GetSelectedItemData(self):
    item = self.protocol_list_widget.selectedItems()[0]
    data = item.data(QtCore.Qt.UserRole) #module_id, in dic_protocol (dictionary)
    index = self.protocol_list_widget.row(item)
    self.details.communicate.SendSelectedItemData([data, index])

  def AddModuleToProtocolWidget(self, module):
    new_module = QListWidgetItem()
    new_module.setText(module["caption"])
    new_module.setData(QtCore.Qt.UserRole, module["id"])
    new_module.setToolTip(module["description"])
    self.protocol_list_widget.addItem(new_module)

  def GetListWidgetModules(self):
    modules_list = []
    for module_iter in range(self.protocol_list_widget.count()):
      module = self.protocol_list_widget.item(module_iter)
      modules_list.append(module)
    self.communicate.ModuleAddedToProtocolLoading(modules_list)

  def SetModuleData(self, module_and_data):
    module = module_and_data[0]
    data = module_and_data[1]
    module.setData(QtCore.Qt.UserRole, data) #new module id



from dtiplayground.dmri.common.tools.base import ExternalToolWrapper

class DTIAverage(ExternalToolWrapper):
    def __init__(self,binary_path= None, **kwargs):
        super().__init__(binary_path, **kwargs)
        self.binary_path=None
        if binary_path is not None:
            self.binary_path=binary_path
        elif 'softwares' in kwargs:
            self.binary_path=kwargs['softwares']['dtiaverage']['path']

    def average(self,input_file_list:list,output_file,options=[]):
        arguments=[]
        for fn in input_file_list:
            arguments+=['--inputs',fn]
        arguments+=['--tensor_output',output_file]
        arguments+=options
        self.setArguments(arguments)
        return self.execute(arguments)
        


#!python

import os
import sys
import argparse
import traceback
from pathlib import Path 
import yaml 
sys.path.append(Path(__file__).parent.__str__())
### for dev

import dtiplayground.dmri.atlasbuilder.utils as utils 
from dtiplayground.dmri.atlasbuilder import AtlasBuilder 
import dtiplayground.dmri.common

logger=dmri.common.logger.write


def build_atlas(args):
    Path(args.output_dir).mkdir(parents=True,exist_ok=True)
    logfilepath=Path(args.output_dir).joinpath('log.txt')
    initialize_logger(logfilepath, args.no_log_timestamp,args.no_verbosity)
    ## system log
    sys_log_dir=current_dir.joinpath('logs')
    sys_log_dir.mkdir(parents=True,exist_ok=True)
    env=os.environ
    uid, ts = dtiplayground.dmri.common.get_uuid(), dtiplayground.dmri.common.get_timestamp()
    sys_logfilename='dmriatlasbuilder_'+env['USER']+"_"+ts+"_"+uid+".txt"
    sys_logfile=sys_log_dir.joinpath(sys_logfilename)
    dtiplayground.dmri.common.logger.addLogfile(sys_logfile.__str__(),mode='w')
    logger("Execution ID : {}".format(uid))
    logger("Execution Command : "+" ".join(sys.argv))
    logger("--------------- Atlasbulder begins ----------------------",dtiplayground.dmri.common.Color.INFO)
    bldr=AtlasBuilder()
    bldr.configure(output_dir=args.output_dir,
                        config_path=args.config,
                        hbuild_path=args.hbuild,
                        greedy_params_path=args.greedy_params,
                        buildsequence_path=args.buildsequence,
                        node=args.node)
   
    bldr.build()

## utilities
def initialize_logger(logpath, no_log_timestamp, no_verbosity):
    ## default log setting
    dtiplayground.dmri.common.logger.setLogfile(logpath)
    dtiplayground.dmri.common.logger.setTimestamp(not no_log_timestamp)
    dtiplayground.dmri.common.logger.setVerbosity(not no_verbosity)

if __name__=="__main__":
    current_dir=Path(__file__).parent
    parser=argparse.ArgumentParser(description="Argument Parser")
    parser.add_argument('--output-dir', help='configuration file',default='output',type=str)
    parser.add_argument('--config',help='configuration file',default='config.yml',type=str)
    parser.add_argument('--hbuild',help='hierarchical build file', default='h-build.yml', type=str)
    parser.add_argument('--greedy-params',help='Greedy atlas parameter xml file', type=str)
    parser.add_argument('--node',help="node to build",type=str)
    parser.add_argument('--buildsequence',help='build sequence file, if this option is inputted then build sequence process will be skipped',type=str)

    ## log related
    parser.add_argument('--no-log-timestamp',help='Remove timestamp in the log', default=False, action="store_true")
    parser.add_argument('--no-verbosity',help='Do not show any logs in the terminal', default=False, action="store_true")

    args=parser.parse_args()

    try:
       logger("--------------- Atlasbulder begins ----------------------",dtiplayground.dmri.common.Color.INFO)
       build_atlas(args)
       sys.exit(0)
    except Exception as e:
       logger(str(e))
       msg=traceback.format_exc()
       logger(msg,dtiplayground.dmri.common.Color.ERROR)
       sys.exit(1)     





#
# Reference : https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3864968/
#
# Written by SK Park , NIRAL, UNC
# 2021-04-18

  
import dtiplayground.dmri.preprocessing as prep

import numpy as np
import time,traceback ,yaml
from pathlib import Path
import os
import markdown
import INTERLACE_Check.computations as computations 

logger=prep.logger.write

class INTERLACE_Check(prep.modules.DTIPrepModule):
    def __init__(self,config_dir,*args,**kwargs):
        super().__init__(config_dir,*args,**kwargs)

    def generateDefaultProtocol(self,image_obj):
        super().generateDefaultProtocol(image_obj)
        ## todos
        spacing=float(np.sum(np.abs(image_obj.information['space_directions']))/3.0)
        self.protocol['translationThreshold']=spacing
        return self.protocol

    def process(self,*args,**kwargs): ## variables : self.source_image, self.image (output) , self.result_history , self.result (output) , self.protocol, self.template
        super().process()
        inputParams=self.getPreviousResult()['output']
        #logger(yaml.dump(inputParams))
        ### Computation 
        output=None
        output_filename=Path(self.computation_dir).joinpath('computations.yml')
        if output_filename.exists() and not self.options['recompute']:
            logger("Recompute : {}".format(self.options['recompute']),prep.Color.INFO)
            logger("There exists the result of interlacing computations",prep.Color.INFO)
            output=yaml.safe_load(open(output_filename,'r'))
            logger("Computed parameters are loaded : {}".format(str(output_filename)),prep.Color.OK)
        else: 
            ### actual computation for interlacing correlation and motions
            logger("Computing interlace correlations and motions ...",prep.Color.PROCESS)
            output=computations.interlace_compute(self.image)
            yaml.dump(output,open(output_filename,'w'))
        ### Check for QC
        logger("Checking bad gradients ...",prep.Color.PROCESS)
        gradient_indexes_to_remove , interlacing_results= computations.interlace_check( self.image,output,
                                                         correlationDeviationBaseline=self.protocol['correlationDeviationBaseline'],
                                                         correlationDeviationGradient=self.protocol['correlationDeviationGradient'],
                                                         correlationThresholdBaseline=self.protocol['correlationThresholdBaseline'],
                                                         correlationThresholdGradient=self.protocol['correlationThresholdGradient'],
                                                         rotationThreshold=self.protocol['rotationThreshold'],
                                                         translationThreshold=self.protocol['translationThreshold'])

        #logger("\nExcluded gradients : {}".format(gradient_indexes_to_remove),prep.Color.WARNING)
        check_filename=Path(self.computation_dir).joinpath('checks.yml')
        yaml.dump(interlacing_results,open(check_filename,'w'))
        logger("Check file saved : {}".format(str(check_filename)),prep.Color.OK)
        ### output preparation
        self.result['output']['excluded_gradients_original_indexes']=self.image.convertToOriginalGradientIndex(gradient_indexes_to_remove)
        self.result['output']['success']=True
        #raise Exception("User Exception for development ...")
        return self.result

    def makeReport(self):
        super().makeReport()        
        
        with open(os.path.abspath(self.output_dir) + '/report.md', 'a') as f:
            if len(self.result['output']['excluded_gradients_original_indexes']) == 0:
                f.write('* 0 excluded gradients\n')
            else:
                excluded_gradients = str(len(self.result['output']['excluded_gradients_original_indexes'])) + " excluded gradient(s): "
                for gradient_index in self.result['output']['excluded_gradients_original_indexes'][:-1]:
                    excluded_gradients = excluded_gradients + str(gradient_index) + ", "
                excluded_gradients += str(self.result['output']['excluded_gradients_original_indexes'][-1])
                f.write('* ' + excluded_gradients + '\n')
        
        self.result['report']['csv_data']['excluded_gradients'] = self.result['output']['excluded_gradients_original_indexes']
        with open(str(Path(self.output_dir).joinpath('result.yml')),'w') as f:
            yaml.dump(self.result,f)


  
import dtiplayground.dmri.preprocessing as prep
import yaml, os
from pathlib import Path
logger=prep.logger.write

import dtiplayground.dmri.common.tools as tools 

###
import SINGLETRACT_Process.tractography as tractography

class SINGLETRACT_Process(prep.modules.DTIPrepModule):
    def __init__(self,config_dir,*args,**kwargs):
        super().__init__(config_dir,*args,**kwargs)
        
    def generateDefaultProtocol(self,image_obj):
        super().generateDefaultProtocol(image_obj)
        p = self.protocol['NIRALUtilitiesPath'].replace('$NIRALUTILS', self.softwares['niral_utilities']['path'])
        self.protocol['NIRALUtilitiesPath'] = p 

        if 'reference_tract' in self.global_variables:
            self.protocol['referenceTractFile'] = Path(self.global_variables['reference_tract']).resolve().__str__()
        ## todos
        return self.protocol
        
    @prep.measure_time
    def process(self,*args,**kwargs): ## variables : self.global_variables, self.softwares, self.output_dir, self.config_dir, self.source_image, self.image (output) , self.result_history , self.result (output) , self.protocol, self.template
        super().process()
        inputParams=self.getPreviousResult()['output']
        protocol_options=args[0]
        self.num_threads=protocol_options['software_info']['parameters']['num_max_threads']
        self.baseline_threshold=protocol_options['baseline_threshold']

        # << TODOS>>
        if self.protocol['method'] == 'NIRAL':
            res = self.singletract_niral(**self.protocol)
        elif self.protocol['method'] == 'dipy':
            res = self.singletract_dipy(**self.protocol)
        else:
            raise Exception("No such method")

        logger(yaml.dump(self.image.information))
        self.result['output']['success']=True

        return self.result

    def singletract_niral(self, **protocol):
        res = {}
        inputDTI = Path(self.output_dir).joinpath('input.nrrd').__str__()
        if 'dti_path' in self.global_variables:
            inputDTI = self.global_variables['dti_path']
        else:
            self.writeImageWithOriginalSpace(inputDTI,'nrrd',dtype='float')
        inputFiberFile = protocol['referenceTractFile']
        displacementField = protocol['displacementFieldFile']
        if not displacementField:
            displacementField = self.global_variables['displacement_field_path']
            if not Path(self.global_variables['displacement_field_path']).exists():
                raise Exception("Couldn't find the displacement field file")

        outputFiberTract = Path(self.output_dir).joinpath('registered_ref_tract.vtk').__str__()
        
    # Register reference tract with the displacement field 
        niralutils = tools.NIRALUtilities(softwares=self.softwares)
        niralutils.dev_mode=True
        arguments = ['--polydata_input', inputFiberFile,
                     '-o', outputFiberTract,
                     '-D', displacementField,
                     '--inverty',
                     '--invertx']
        if self.overwriteFile(outputFiberTract) : output=niralutils.polydatatransform(arguments)
        self.addGlobalVariable('reference_tract_path', outputFiberTract) ## update tract path to the updated one

    ## Dilation and voxelization of the mapped reference tracts , getting labelmap
        labelMapFile = Path(self.output_dir).joinpath('labelmap.nrrd').__str__()
        arguments = ['--voxelize', labelMapFile,
                     '--fiber_file', outputFiberTract,
                     '-T', inputDTI]
        fiberprocess = tools.FiberProcess(softwares=self.softwares)
        fiberprocess.dev_mode = True
        if self.overwriteFile(labelMapFile) : fiberprocess.execute(arguments=arguments)
        self.addGlobalVariable('labelmap_path', labelMapFile)

    ## Dilation and voxelization of the reference tracts
        dilatedLabelmapFile = Path(self.output_dir).joinpath('labelmap_dilated.nrrd').__str__()
        dilationRadius = self.protocol['dilationRadius']
        imagemath = tools.ImageMath(softwares=self.softwares)
        imagemath.dev_mode=True
        arguments = [labelMapFile,
                     '-dilate', str(dilationRadius)+',1',
                     '-outfile', dilatedLabelmapFile
                     ]
        if self.overwriteFile(dilatedLabelmapFile) : imagemath.execute(arguments=arguments)
        self.addGlobalVariable('labelmap_path', dilatedLabelmapFile)
        labelMapImage = prep.dwi.DWI(labelMapFile)
        dilatedLabelmapImage = prep.dwi.DWI(dilatedLabelmapFile)
        dipyLabelMap = labelMapImage.images + dilatedLabelmapImage.images

    ## Tractography ...
        tensorImage = prep.dwi.DWI(inputDTI)
        tractography.compute(tensorImage, self.image, dipyLabelMap, self.output_dir )
        return res 


    def singletract_dipy(self, **protocol):
        res = {}
        return res 



from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
from PyQt5.QtCore import * 

from dtiplayground.ui.modules_details_communicate import ModulesDetailsCommunicate

class QCReport(QWidget):
  communicate = ModulesDetailsCommunicate()

  def __init__(self, protocol_template, qcreport_yml):
    QWidget.__init__(self)
    self.stack = QWidget()
    self.QCReportStack(protocol_template, qcreport_yml)

  def QCReportStack(self, protocol_template, qcreport_yml):

    ## Module
    self.tab_name = QLabel()

    ## Description
    for template_ite in range(len(protocol_template["options"]["execution"]["pipeline"]["candidates"])):
      if protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["value"] == "QC_Report":
        description_label = QLabel(protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["description"])
    description_label.setWordWrap(True)

    ## Options:    
    options_groupbox = QGroupBox("Options:")
    options_layout = QGridLayout()
    options_groupbox.setLayout(options_layout)    
    # overwrite
    self.overwrite = QCheckBox("Overwrite")    
    self.overwrite.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.overwrite, 0, 0)
    # skip
    self.skip = QCheckBox("Skip")
    self.skip.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.skip, 0, 1)
    # write image
    self.writeimage = QCheckBox("Write Image")
    
    ## Protocol:
    protocol_groupbox = QGroupBox("Protocol:")
    protocol_layout = QFormLayout()
    protocol_groupbox.setLayout(protocol_layout)
    # generatePDF
    self.generatePDF_true = QRadioButton("True")
    self.generatePDF_false = QRadioButton("False")
    generatePDF_layout = QHBoxLayout()
    generatePDF_layout.addWidget(self.generatePDF_true)
    generatePDF_layout.addWidget(self.generatePDF_false)
    self.generatePDF_buttongroup = QButtonGroup()
    self.generatePDF_buttongroup.addButton(self.generatePDF_true)
    self.generatePDF_buttongroup.addButton(self.generatePDF_false)
    self.generatePDF_true.toggled.connect(self.GetParams)
    self.generatePDF_true.setStatusTip(qcreport_yml["protocol"]["generatePDF"]["description"])
    self.generatePDF_false.setStatusTip(qcreport_yml["protocol"]["generatePDF"]["description"])
    protocol_layout.addRow(qcreport_yml["protocol"]["generatePDF"]["caption"], generatePDF_layout)
    # generateCSV
    self.generateCSV_true = QRadioButton("True")
    self.generateCSV_false = QRadioButton("False")
    generateCSV_layout = QHBoxLayout()
    generateCSV_layout.addWidget(self.generateCSV_true)
    generateCSV_layout.addWidget(self.generateCSV_false)
    self.generateCSV_buttongroup = QButtonGroup()
    self.generateCSV_buttongroup.addButton(self.generateCSV_true)
    self.generateCSV_buttongroup.addButton(self.generateCSV_false)
    self.generateCSV_true.toggled.connect(self.GetParams)
    self.generateCSV_true.setStatusTip(qcreport_yml["protocol"]["generateCSV"]["description"])
    self.generateCSV_false.setStatusTip(qcreport_yml["protocol"]["generateCSV"]["description"])
    protocol_layout.addRow(qcreport_yml["protocol"]["generateCSV"]["caption"], generateCSV_layout)    
    
    ## Layout
    layout_v = QVBoxLayout()
    layout_v.addWidget(self.tab_name)
    layout_v.addWidget(description_label)
    layout_v.addWidget(options_groupbox)
    layout_v.addWidget(protocol_groupbox)
    layout_v.addStretch(1)
    self.stack.setLayout(layout_v)


  def GetParams(self):
    params = [
      'QC Report',
      'QC_Report', 
      {'options': {
        'overwrite': self.overwrite.isChecked(), 
        'skip': self.skip.isChecked(), 
        }, 
      'protocol': {
        'generatePDF': self.generatePDF_true.isChecked(), 
        'generateCSV': self.generateCSV_true.isChecked()
        }
      }
    ]
    self.communicate.SendParams(params)


from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
from PyQt5.QtCore import * 

from dtiplayground.ui.modules_details_communicate import ModulesDetailsCommunicate

class DTIEstimate(QWidget):
  communicate = ModulesDetailsCommunicate()

  def __init__(self, protocol_template, dtiestimate_yml):
    QWidget.__init__(self)
    self.dtiestimate_yml = dtiestimate_yml
    self.stack = QWidget()
    self.DTIEstimateStack(protocol_template)

  def DTIEstimateStack(self, protocol_template):

    ## Module
    self.tab_name = QLabel()

    ## Description
    for template_ite in range(len(protocol_template["options"]["execution"]["pipeline"]["candidates"])):
      if protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["value"] == "DTI_Estimate":
        description_label = QLabel(protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["description"])
    description_label.setWordWrap(True)

    ## Options:
    options_groupbox = QGroupBox("Options:")
    options_layout = QGridLayout()
    options_groupbox.setLayout(options_layout)   
    # overwrite
    self.overwrite = QCheckBox("Overwrite")
    self.overwrite.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.overwrite, 0, 0)
    # skip
    self.skip = QCheckBox("Skip")
    self.skip.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.skip, 0, 1)
    # write image
    self.writeimage = QCheckBox("Write Image")
    self.writeimage.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.writeimage, 0, 2)
    
    ## Protocol:
    protocol_groupbox = QGroupBox("Protocol:")
    protocol_layout = QFormLayout()
    protocol_groupbox.setLayout(protocol_layout)
    # method
    self.method = QComboBox()
    for ite in self.dtiestimate_yml["protocol"]["method"]["candidates"]:
      self.method.addItem(ite["caption"])
      self.method.setItemData(self.method.count()-1, ite["description"], QtCore.Qt.ToolTipRole)
    self.method.currentTextChanged.connect(self.GetMethodIt)
    self.method.setStatusTip(self.dtiestimate_yml["protocol"]["method"]["description"])
    protocol_layout.addRow(self.dtiestimate_yml["protocol"]["method"]["caption"], self.method)
    self.method.currentTextChanged.connect(self.EnableDisableMethods)
    self.method.currentTextChanged.connect(self.GetParams)
    # optimizationMethod
    self.optimizationMethod = QComboBox()
    for ite in self.dtiestimate_yml["protocol"]["optimizationMethod"]["candidates"]:
      self.optimizationMethod.addItem(ite["caption"])
    self.optimizationMethod.currentTextChanged.connect(self.GetOptimizationMethodIt)
    #self.optimizationMethod.setStatusTip(self.dtiestimate_yml["protocol"]["optimizationMethod"]["description"])
    protocol_layout.addRow(self.dtiestimate_yml["protocol"]["optimizationMethod"]["caption"], self.optimizationMethod)
    self.optimizationMethod.currentTextChanged.connect(self.GetParams)
    # correctionMethod
    self.correctionMethod = QComboBox()
    for ite in self.dtiestimate_yml["protocol"]["correctionMethod"]["candidates"]:
      self.correctionMethod.addItem(ite["caption"])
      self.correctionMethod.setItemData(self.correctionMethod.count()-1, ite["description"], QtCore.Qt.ToolTipRole)
    self.correctionMethod.currentTextChanged.connect(self.GetCorrectionMethodIt)
    self.correctionMethod.setStatusTip(self.dtiestimate_yml["protocol"]["correctionMethod"]["description"])
    protocol_layout.addRow(self.dtiestimate_yml["protocol"]["correctionMethod"]["caption"], self.correctionMethod)
    self.correctionMethod.currentTextChanged.connect(self.GetParams)

    ## Layout
    layout_v = QVBoxLayout()
    layout_v.addWidget(self.tab_name)
    layout_v.addWidget(description_label)
    layout_v.addWidget(options_groupbox)
    layout_v.addWidget(protocol_groupbox)
    layout_v.addStretch(1)
    self.stack.setLayout(layout_v)

    self.EnableDisableMethods()

  def EnableDisableMethods(self): # enables and disables optimization methods and correction methods according to the selected method
    if self.method.currentText() == "DIPY":
      self.optimizationMethod.model().item(self.optimizationMethod.findText("RESTORE (DIPY Only)", Qt.MatchExactly)).setEnabled(True)
      self.optimizationMethod.model().item(self.optimizationMethod.findText("Maximum Likelihood (dtiestim Only)", Qt.MatchExactly)).setEnabled(False)
      self.correctionMethod.model().item(self.correctionMethod.findText("Nearest (dtiestim only)", Qt.MatchExactly)).setEnabled(False)
      self.correctionMethod.model().item(self.correctionMethod.findText("Absolute (dtiestim only)", Qt.MatchExactly)).setEnabled(False)
      self.correctionMethod.model().item(self.correctionMethod.findText("None (dtiestim only)", Qt.MatchExactly)).setEnabled(False)
      if self.correctionMethod.currentText() != "Zero":
        self.correctionMethod.setCurrentText("Zero")
      if self.optimizationMethod.currentText() == "Maximum Likelihood (dtiestim Only)":
        self.optimizationMethod.setCurrentText("Weighted Least Squares")
    if self.method.currentText() == "dtiestim":
      self.optimizationMethod.model().item(self.optimizationMethod.findText("RESTORE (DIPY Only)", Qt.MatchExactly)).setEnabled(False)
      self.optimizationMethod.model().item(self.optimizationMethod.findText("Maximum Likelihood (dtiestim Only)", Qt.MatchExactly)).setEnabled(True)
      self.correctionMethod.model().item(self.correctionMethod.findText("Nearest (dtiestim only)", Qt.MatchExactly)).setEnabled(True)
      self.correctionMethod.model().item(self.correctionMethod.findText("Absolute (dtiestim only)", Qt.MatchExactly)).setEnabled(True)
      self.correctionMethod.model().item(self.correctionMethod.findText("None (dtiestim only)", Qt.MatchExactly)).setEnabled(True)
      if self.optimizationMethod.currentText() == "RESTORE (DIPY Only)":
        self.optimizationMethod.setCurrentText("Weighted Least Squares")

  def GetMethodIt(self, text): # dti estimate parameter
    self.method_it = 0
    for it in range(len(self.dtiestimate_yml["protocol"]["method"]["candidates"])):
      if self.dtiestimate_yml["protocol"]["method"]["candidates"][it]["caption"] == text:
        self.method_it = it

  def GetOptimizationMethodIt(self, text): # dti estimate parameter
    self.optimizationmethod_it = 0
    for it in range(len(self.dtiestimate_yml["protocol"]["optimizationMethod"]["candidates"])):
      if self.dtiestimate_yml["protocol"]["optimizationMethod"]["candidates"][it]["caption"] == text:
        self.optimizationmethod_it = it

  def GetCorrectionMethodIt(self, text): # dti estimate parameter
    self.correctionmethod_it = 0
    for it in range(len(self.dtiestimate_yml["protocol"]["correctionMethod"]["candidates"])):
      if self.dtiestimate_yml["protocol"]["correctionMethod"]["candidates"][it]["caption"] == text:
        self.correctionmethod_it = it

  def GetParams(self):
    params = [
      'Estimate DTI',
      'DTI_Estimate', 
      {'options': {
        'overwrite': self.overwrite.isChecked(), 
        'skip': self.skip.isChecked(), 
        'write_image': self.writeimage.isChecked()
        }, 
      'protocol': {
        'correctionMethod': self.dtiestimate_yml["protocol"]["correctionMethod"]["candidates"][self.correctionmethod_it]["value"], 
        'method': self.dtiestimate_yml["protocol"]["method"]["candidates"][self.method_it]["value"],
        'optimizationMethod': self.dtiestimate_yml["protocol"]["optimizationMethod"]["candidates"][self.optimizationmethod_it]["value"],        
        }
      }
    ]
    self.communicate.SendParams(params)


from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
from PyQt5.QtCore import * 

from dtiplayground.ui.modules_details_communicate import ModulesDetailsCommunicate

class EddyMotion(QWidget):
  communicate = ModulesDetailsCommunicate()

  def __init__(self, protocol_template, eddymotion_yml):
    QWidget.__init__(self)
    self.stack = QWidget()
    self.EddyMotionStack(protocol_template, eddymotion_yml)

  def EddyMotionStack(self, protocol_template, eddymotion_yml):

    ## Module
    self.tab_name = QLabel()

    ## Description
    for template_ite in range(len(protocol_template["options"]["execution"]["pipeline"]["candidates"])):
      if protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["value"] == "EDDYMOTION_Correct":
        description_label = QLabel(protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["description"])
    description_label.setWordWrap(True)
    
    ## Options:
    options_groupbox = QGroupBox("Options:")
    options_layout = QGridLayout()
    options_groupbox.setLayout(options_layout)   
    # overwrite
    self.overwrite = QCheckBox("Overwrite")
    self.overwrite.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.overwrite, 0, 0)
    # skip
    self.skip = QCheckBox("Skip")
    self.skip.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.skip, 0, 1)
    # write image
    self.writeimage = QCheckBox("Write Image")
    self.writeimage.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.writeimage, 0, 2)

    ## Protocol:
    protocol_groupbox = QGroupBox("Protocol:")
    protocol_layout = QFormLayout()
    protocol_groupbox.setLayout(protocol_layout)
    # estimateMoveBySusceptibility
    self.estimateMoveBySusceptibility_true = QRadioButton("True")
    self.estimateMoveBySusceptibility_false = QRadioButton("False")
    estimateMoveBySusceptibility_layout = QHBoxLayout()
    estimateMoveBySusceptibility_layout.addWidget(self.estimateMoveBySusceptibility_true)
    estimateMoveBySusceptibility_layout.addWidget(self.estimateMoveBySusceptibility_false)
    self.estimateMoveBySusceptibility_buttongroup = QButtonGroup()
    self.estimateMoveBySusceptibility_buttongroup.addButton(self.estimateMoveBySusceptibility_true)
    self.estimateMoveBySusceptibility_buttongroup.addButton(self.estimateMoveBySusceptibility_false)
    self.estimateMoveBySusceptibility_true.toggled.connect(self.GetParams)
    self.estimateMoveBySusceptibility_true.setStatusTip(eddymotion_yml["protocol"]["estimateMoveBySusceptibility"]["description"])
    self.estimateMoveBySusceptibility_false.setStatusTip(eddymotion_yml["protocol"]["estimateMoveBySusceptibility"]["description"])
    protocol_layout.addRow(eddymotion_yml["protocol"]["estimateMoveBySusceptibility"]["caption"], estimateMoveBySusceptibility_layout)
    self.estimateMoveBySusceptibility_true.setEnabled(False)
    self.estimateMoveBySusceptibility_false.setEnabled(False)
    # interpolateBadData
    self.interpolateBadData_true = QRadioButton("True")
    self.interpolateBadData_false = QRadioButton("False")
    interpolateBadData_layout = QHBoxLayout()
    interpolateBadData_layout.addWidget(self.interpolateBadData_true)
    interpolateBadData_layout.addWidget(self.interpolateBadData_false)
    self.interpolateBadData_buttongroup = QButtonGroup()
    self.interpolateBadData_buttongroup.addButton(self.interpolateBadData_true)
    self.interpolateBadData_buttongroup.addButton(self.interpolateBadData_false)
    self.interpolateBadData_true.toggled.connect(self.GetParams)
    self.interpolateBadData_true.setStatusTip(eddymotion_yml["protocol"]["interpolateBadData"]["description"])
    self.interpolateBadData_false.setStatusTip(eddymotion_yml["protocol"]["interpolateBadData"]["description"])
    protocol_layout.addRow(eddymotion_yml["protocol"]["interpolateBadData"]["caption"], interpolateBadData_layout)
    # dataIsShelled
    self.dataIsShelled_true = QRadioButton("True")
    self.dataIsShelled_false = QRadioButton("False")
    dataIsShelled_layout = QHBoxLayout()
    dataIsShelled_layout.addWidget(self.dataIsShelled_true)
    dataIsShelled_layout.addWidget(self.dataIsShelled_false)
    self.dataIsShelled_buttongroup = QButtonGroup()
    self.dataIsShelled_buttongroup.addButton(self.dataIsShelled_true)
    self.dataIsShelled_buttongroup.addButton(self.dataIsShelled_false)
    self.dataIsShelled_true.toggled.connect(self.GetParams)
    self.dataIsShelled_true.setStatusTip(eddymotion_yml["protocol"]["dataIsShelled"]["description"])
    self.dataIsShelled_false.setStatusTip(eddymotion_yml["protocol"]["dataIsShelled"]["description"])
    protocol_layout.addRow(eddymotion_yml["protocol"]["dataIsShelled"]["caption"], dataIsShelled_layout)
    # qcReport
    self.qcReport_true = QRadioButton("True")
    self.qcReport_false = QRadioButton("False")
    qcReport_layout = QHBoxLayout()
    qcReport_layout.addWidget(self.qcReport_true)
    qcReport_layout.addWidget(self.qcReport_false)
    self.qcReport_buttongroup = QButtonGroup()
    self.qcReport_buttongroup.addButton(self.qcReport_true)
    self.qcReport_buttongroup.addButton(self.qcReport_false)
    self.qcReport_true.toggled.connect(self.GetParams)
    self.qcReport_true.setStatusTip(eddymotion_yml["protocol"]["qcReport"]["description"])
    self.qcReport_false.setStatusTip(eddymotion_yml["protocol"]["qcReport"]["description"])
    protocol_layout.addRow(eddymotion_yml["protocol"]["qcReport"]["caption"], qcReport_layout)

    ## Layout
    layout_v = QVBoxLayout()
    layout_v.addWidget(self.tab_name)
    layout_v.addWidget(description_label)
    layout_v.addWidget(options_groupbox)
    layout_v.addWidget(protocol_groupbox)
    layout_v.addStretch(1)
    self.stack.setLayout(layout_v)

  def GetParams(self):
    params = [
      'Eddy motion Correction',
      'EDDYMOTION_Correct', 
      {'options': {
        'overwrite': self.overwrite.isChecked(), 
        'skip': self.skip.isChecked(), 
        'write_image': self.writeimage.isChecked()
        }, 
      'protocol': {
        'estimateMoveBySusceptibility': self.estimateMoveBySusceptibility_true.isChecked(),
        'interpolateBadData': self.interpolateBadData_true.isChecked(),
        'dataIsShelled': self.dataIsShelled_true.isChecked(),
        'qcReport': self.qcReport_true.isChecked()
        }
      }
    ]
    self.communicate.SendParams(params)


from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
from PyQt5.QtCore import * 

from dtiplayground.ui.modules_details_communicate import ModulesDetailsCommunicate

class BrainMask(QWidget):
  communicate = ModulesDetailsCommunicate()

  def __init__(self, protocol_template, brainmask_yml):
    QWidget.__init__(self)
    self.brainmask_yml = brainmask_yml
    self.method_it = 0
    self.averagingmethod_it = 0
    self.stack = QWidget()
    self.BrainMaskStack(protocol_template)

  def BrainMaskStack(self, protocol_template):

    ## Module
    self.tab_name = QLabel()
    
    ## Description
    for template_ite in range(len(protocol_template["options"]["execution"]["pipeline"]["candidates"])):
      if protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["value"] == "BRAIN_Mask":
        description_label = QLabel(protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["description"])
    description_label.setWordWrap(True)

    ## Options:
    options_groupbox = QGroupBox("Options:")
    options_layout = QGridLayout()
    options_groupbox.setLayout(options_layout)   
    # overwrite
    self.overwrite = QCheckBox("Overwrite")
    self.overwrite.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.overwrite, 0, 0)
    # skip
    self.skip = QCheckBox("Skip")
    self.skip.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.skip, 0, 1)
    # write image
    self.writeimage = QCheckBox("Write Image")

    ## Protocol:
    protocol_groupbox = QGroupBox("Protocol:")
    protocol_layout = QVBoxLayout()
    protocol_sublayout1 = QFormLayout()
    protocol_sublayout2 = QHBoxLayout()
    protocol_groupbox.setLayout(protocol_layout)
    # method
    self.method = QComboBox()
    for ite in self.brainmask_yml["protocol"]["method"]["candidates"]:
      self.method.addItem(ite["caption"])
      self.method.setItemData(self.method.count()-1, ite["description"], QtCore.Qt.ToolTipRole)
    self.method.currentTextChanged.connect(self.GetMethodIt)
    self.method.setStatusTip(self.brainmask_yml["protocol"]["method"]["description"])
    protocol_sublayout1.addRow(self.brainmask_yml["protocol"]["method"]["caption"], self.method)
    self.method.currentTextChanged.connect(self.GetParams)
    # averaging method
    self.averagingmethod = QComboBox()
    for ite in self.brainmask_yml["protocol"]["averagingMethod"]["candidates"]:
      self.averagingmethod.addItem(ite["caption"])
      self.averagingmethod.setItemData(self.averagingmethod.count()-1, ite["description"], QtCore.Qt.ToolTipRole)
    self.averagingmethod.currentTextChanged.connect(self.GetAveragingMethodIt)
    self.averagingmethod.setStatusTip(self.brainmask_yml["protocol"]["averagingMethod"]["description"])
    protocol_sublayout1.addRow(self.brainmask_yml["protocol"]["averagingMethod"]["caption"], self.averagingmethod)
    self.averagingmethod.currentTextChanged.connect(self.GetParams)
    # modality
    self.modality_label = QLabel("Modality")
    self.modality_t2 = QRadioButton("T2")
    self.modality_fa = QRadioButton("FA")
    modality_buttongroup = QButtonGroup()
    modality_buttongroup.addButton(self.modality_t2)
    modality_buttongroup.addButton(self.modality_fa)
    #protocol_sublayout2.addWidget(self.modality_label)
    #protocol_sublayout2.addWidget(self.modality_fa)
    #protocol_sublayout2.addWidget(self.modality_t2)
    #self.method.currentTextChanged.connect(self.DisplayModality)  
    #self.GetModalityIt()
    self.modality_t2.toggled.connect(self.GetModalityIt)
    self.modality_t2.toggled.connect(self.GetParams)
    #self.modality_label.setStatusTip(self.brainmask_yml["protocol"]["modality"]["description"])
    #self.modality_t2.setToolTip(self.brainmask_yml["protocol"]["modality"]["candidates"][0]["description"])
    #self.modality_fa.setToolTip(self.brainmask_yml["protocol"]["modality"]["candidates"][1]["description"])

    ## Layout
    protocol_layout.addLayout(protocol_sublayout1)
    protocol_layout.addLayout(protocol_sublayout2)
    layout_v = QVBoxLayout()
    layout_v.addWidget(self.tab_name)
    layout_v.addWidget(description_label)
    layout_v.addWidget(options_groupbox)
    layout_v.addWidget(protocol_groupbox)
    layout_v.addStretch(1)
    self.stack.setLayout(layout_v)

  def GetMethodIt(self, text): # get "method" parameter
    self.method_it = 0
    for it in range(len(self.brainmask_yml["protocol"]["method"]["candidates"])):
      if self.brainmask_yml["protocol"]["method"]["candidates"][it]["caption"] == text:
        self.method_it = it

  def GetAveragingMethodIt(self, text): # get "average method" parameter
    for it in range(len(self.brainmask_yml["protocol"]["averagingMethod"]["candidates"])):
      if self.brainmask_yml["protocol"]["averagingMethod"]["candidates"][it]["caption"] == text:
        self.averagingmethod_it = it

  def GetModalityIt(self): # get "modality" parameter
    self.modality_it = 0
    if self.modality_t2.isChecked():
      self.modality_it = 0
    if self.modality_fa.isChecked():
      self.modality_it = 1

  def DisplayModality(self):
    if self.method.currentText() != "AntsPyNet":
      self.modality_t2.setHidden(True)
      self.modality_fa.setHidden(True)
      self.modality_label.setHidden(True)
    else:
      self.modality_t2.setHidden(False)
      self.modality_fa.setHidden(False)
      self.modality_label.setHidden(False)

  def GetParams(self):
    params = [
      'Brain Masking',
      'BRAIN_Mask', 
      {'options': {
        'overwrite': self.overwrite.isChecked(), 
        'skip': self.skip.isChecked(), 
        }, 
      'protocol': {
        'averagingMethod': self.brainmask_yml["protocol"]["averagingMethod"]["candidates"][self.averagingmethod_it]["value"],
        'method': self.brainmask_yml["protocol"]["method"]["candidates"][self.method_it]["value"], 
        #'modality': self.brainmask_yml["protocol"]["modality"]["candidates"][self.modality_it]["value"]
        }
      }
    ]
    self.communicate.SendParams(params)


from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
from PyQt5.QtCore import * 
import yaml
from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal as Signal


class ProtocolCommunicate(QObject):

  call_clear_protocol = Signal()
  set_general_param_from_read_protocol = Signal(dict)
  add_module_to_protocol_widget = Signal(dict)
  get_protocol_io = Signal(list)
  call_unsaved_modif = Signal()
  get_index_and_key_to_update_module_details = Signal()
  get_list_widget_modules = Signal()
  get_selected_modules = Signal()
  set_module_data = Signal(list)
  call_execute_exclude_gradients_popup = Signal()
  call_execute_merge_images_popup = Signal()
  enable_modules_list_widget = Signal(bool)
  enable_exclude_gradients_module = Signal(bool)
  enable_merge_images_module = Signal(bool)
  update_selector_data = Signal(str, int)
  enable_eddymotion_param = Signal(bool)

  def CallClearProtocol(self):
    self.call_clear_protocol.emit()

  def SetGeneralParamFromReadProtocol(self, protocol_io):
    self.set_general_param_from_read_protocol.emit(protocol_io)

  def AddModuleToProtocolWidget(self, module):
    self.add_module_to_protocol_widget.emit(module)

  def GetProtocolIO(self, list):
    self.get_protocol_io.emit(list)

  def CallUnsavedModif(self):
    self.call_unsaved_modif.emit()
  
  def GetListWidgetModules(self):
    self.get_list_widget_modules.emit()

  def GetSelectedModules(self):
    self.get_selected_modules.emit()

  def SetModuleData(self, module_and_data):
    self.set_module_data.emit(module_and_data)
  
  def CallExecuteExcludeGradientsPopup(self):
    self.call_execute_exclude_gradients_popup.emit()

  def CallExecuteMergeImagesPopup(self):
    self.call_execute_merge_images_popup.emit()

  def EnableModulesListWidget(self, new_state):
    self.enable_modules_list_widget.emit(new_state)

  def EnableExcludeGradientsModule(self, new_state):
    self.enable_exclude_gradients_module.emit(new_state)

  def EnableMergeImagesModule(self, new_state):
    self.enable_merge_images_module.emit(new_state)

  def UpdateSelectorData(self, module_name, new_data):
    self.update_selector_data.emit(module_name, new_data)

  def EnableEddyMotionParam(self, value):
    self.enable_eddymotion_param.emit(value)

class Protocol():
  communicate = ProtocolCommunicate()

  def __init__(self, protocol_template, protocol_yml, preferences_yml, init_module_id):
    self.protocol_template = protocol_template
    self.preferences_yml = preferences_yml
    self.module_id = init_module_id
    self.dic_protocol = {}
    self.loading_protocol = False
    self.update_module_details = True
    self.InitializeModulesParameters(protocol_yml)

  def ReadProtocol(self, protocol):
    print("Reading protocol", protocol)
    self.communicate.CallClearProtocol()
    self.loading_protocol = True

    protocol_yml = yaml.safe_load(open(protocol,'r'))

    self.communicate.SetGeneralParamFromReadProtocol(protocol_yml["io"])    

    for protocol_module in protocol_yml["pipeline"]:
      if protocol_module[0] == "SLICE_Check":
        self.slicecheck_params[2] = protocol_module[1]
      if protocol_module[0] == "INTERLACE_Check":
        self.interlacecheck_params[2] = protocol_module[1]
      if protocol_module[0] == "BASELINE_Average":
        self.baselineaverage_params[2] = protocol_module[1]
      if protocol_module[0] == "SUSCEPTIBILITY_Correct":
        self.susceptibility_params[2] = protocol_module[1]
      if protocol_module[0] == "EDDYMOTION_Correct":
        self.eddymotion_params[2] = protocol_module[1]
      if protocol_module[0] == "BRAIN_Mask":
        self.brainmask_params[2] = protocol_module[1]
      if protocol_module[0] == "DTI_Estimate":
        self.dtiestimate_params[2] = protocol_module[1]
      if protocol_module[0] == "MANUAL_Exclude":
        self.exclude_params[2] = protocol_module[1]
      if protocol_module[0] == "UTIL_Header":
        self.utilheader_params[2] = protocol_module[1]
      if protocol_module[0] == "UTIL_Merge":
        self.utilmerge_params[2] = protocol_module[1]
      if protocol_module[0] == "QC_Report":
        self.qcreport_params[2] = protocol_module[1]

      # modules list
      for template_module in self.protocol_template["options"]["execution"]["pipeline"]["candidates"]:
        if template_module["value"] == protocol_module[0]:
          template_module["id"] = self.module_id
          self.communicate.AddModuleToProtocolWidget(template_module)

      self.module_id += 1
      
    self.loading_protocol = False

  def WriteProtocolYML(self, name):
    protocol = {
      'io': None, 
      'pipeline': [
      ]
    }
    
    for module in self.dic_protocol.keys():
      protocol["pipeline"].append(self.dic_protocol[module][1:])

    protocol["version"] = self.protocol_template["version"]

    data = [protocol, name]
    self.communicate.GetProtocolIO(data)

  def Writing(self, data):
    protocol = data[0]
    name = data[1]
    with open(name, 'w') as protocolfile:
      yaml.dump(protocol, protocolfile)

  def OrderModulesInProtocolList(self, list_new_keys):
    new_dic_protocol = {}
    for key in list_new_keys:
      new_dic_protocol[key] = self.dic_protocol[key]
    self.SetDicProtocol(new_dic_protocol)

  def ModuleAddedToProtocol(self):    
    if self.loading_protocol == True:
      self.communicate.GetListWidgetModules()      
    else:
      self.communicate.GetSelectedModules()
    self.communicate.CallUnsavedModif()

  def ModuleAddedToProtocolLoading(self, list_modules):
    exclude_in_protocol = False
    merge_in_protocol = False
    for module in list_modules:
      key = module.data(QtCore.Qt.UserRole)
      if module.text() == "Exclude Gradients":
        exclude_in_protocol = True
        self.dic_protocol[key] = self.exclude_params
      if module.text() == "Susceptibility correction":
        self.communicate.EnableEddyMotionParam(True)
        self.dic_protocol[key] = self.susceptibility_params
      if module.text() == "Slicewise Check":
        self.dic_protocol[key] = self.slicecheck_params
      if module.text() == "Interlace Correlation Check":
        self.dic_protocol[key] = self.interlacecheck_params
      if module.text() == "Baseline Average":
        self.dic_protocol[key] = self.baselineaverage_params
      if module.text() == "Eddy motion Correction":
        self.dic_protocol[key] = self.eddymotion_params
      if module.text() == "Brain Masking":
        self.dic_protocol[key] = self.brainmask_params
      if module.text() == "Estimate DTI":
        self.dic_protocol[key] = self.dtiestimate_params
      if module.text() == "View Header":
        self.dic_protocol[key] = self.utilheader_params
      if module.text() == "Merge Images":
        self.dic_protocol[key] = self.utilmerge_params
        merge_in_protocol = True
      if module.text() == "QC Report":
        self.dic_protocol[key] = self.qcreport_params

    self.CheckSpecialModulesInProtocol(exclude_in_protocol, merge_in_protocol)

  def ModuleAddedToProtocolDrop(self, list_selected_modules):
    exclude_in_protocol = False
    merge_in_protocol = False
    for module in list_selected_modules:
      current_id = module.data(QtCore.Qt.UserRole)
      if module.text() == "Exclude Gradients":
        exclude_in_protocol = True
        self.dic_protocol[current_id] = self.exclude_default_params
      if module.text() == "Susceptibility correction":
        self.dic_protocol[current_id] = self.susceptibility_default_params
        self.communicate.EnableEddyMotionParam(True)
      if module.text() == "Slicewise Check":
        self.dic_protocol[current_id] = self.slicecheck_default_params
      if module.text() == "Interlace Correlation Check":
        self.dic_protocol[current_id] = self.interlacecheck_default_params
      if module.text() == "Baseline Average":
        self.dic_protocol[current_id] = self.baselineaverage_default_params
      if module.text() == "Eddy motion Correction":
        self.dic_protocol[current_id] = self.eddymotion_default_params
      if module.text() == "Brain Masking":
        self.dic_protocol[current_id] = self.brainmask_default_params
      if module.text() == "Estimate DTI":
        self.dic_protocol[current_id] = self.dtiestimate_default_params
      if module.text() == "View Header":
        self.dic_protocol[current_id] = self.utilheader_default_params
      if module.text() == "Merge Images":
        self.dic_protocol[current_id] = self.utilmerge_default_params
        merge_in_protocol = True
      if module.text() == "QC Report":
        self.dic_protocol[current_id] = self.qcreport_default_params
      
      self.communicate.SetModuleData([module, current_id])
      self.module_id += 1
      self.communicate.UpdateSelectorData(module.text(), self.module_id)

    self.CheckSpecialModulesInProtocol(exclude_in_protocol, merge_in_protocol)

  def CheckSpecialModulesInProtocol(self, exclude_in_protocol, merge_in_protocol):
    if exclude_in_protocol:
      if self.preferences_yml["showExcludeGradientsPopup"] == True:
        self.communicate.CallExecuteExcludeGradientsPopup()
      self.communicate.EnableModulesListWidget(False)
    elif merge_in_protocol:
      if self.preferences_yml["showMergeImagesPopup"] == True:
        self.communicate.CallExecuteMergeImagesPopup()
      self.communicate.EnableModulesListWidget(False)
    else:
      self.communicate.EnableModulesListWidget(True)
      self.communicate.EnableMergeImagesModule(False)
      self.communicate.EnableExcludeGradientsModule(False)

    
  def InitializeModulesParameters(self, protocol_yml):
    for ite in range(len(protocol_yml["pipeline"])):

      if protocol_yml["pipeline"][ite][0] == "SLICE_Check":
        self.slicecheck_default_params = protocol_yml["pipeline"][ite]
        self.slicecheck_default_params.insert(0, "Slicewise Check")

      if protocol_yml["pipeline"][ite][0] == "INTERLACE_Check":
        self.interlacecheck_default_params = protocol_yml["pipeline"][ite]
        self.interlacecheck_default_params.insert(0, "Interlace Correlation Check")
        
      if protocol_yml["pipeline"][ite][0] == "BASELINE_Average":
        self.baselineaverage_default_params = protocol_yml["pipeline"][ite]
        self.baselineaverage_default_params.insert(0, "Baseline Average")
        
      if protocol_yml["pipeline"][ite][0] == "SUSCEPTIBILITY_Correct":
        protocol_yml["pipeline"][ite][1]["protocol"]["configurationFilePath"] = self.preferences_yml["fslConfigurationFilePath"]
        self.susceptibility_default_params = protocol_yml["pipeline"][ite]
        self.susceptibility_default_params.insert(0, "Susceptibility correction")
        
      if protocol_yml["pipeline"][ite][0] == "EDDYMOTION_Correct":
        self.eddymotion_default_params = protocol_yml["pipeline"][ite]
        self.eddymotion_default_params.insert(0, "Eddy motion Correction")
        
      if protocol_yml["pipeline"][ite][0] == "BRAIN_Mask":
        self.brainmask_default_params = protocol_yml["pipeline"][ite]
        self.brainmask_default_params.insert(0, "Brain Masking")

      if protocol_yml["pipeline"][ite][0] == "DTI_Estimate":
        self.dtiestimate_default_params = protocol_yml["pipeline"][ite]
        self.dtiestimate_default_params.insert(0, "Estimate DTI")

      if protocol_yml["pipeline"][ite][0] == "MANUAL_Exclude":
        self.exclude_default_params = protocol_yml["pipeline"][ite]
        self.exclude_default_params.insert(0, "Exclude Gradients")
        
      if protocol_yml["pipeline"][ite][0] == "UTIL_Header":
        self.utilheader_default_params = protocol_yml["pipeline"][ite]
        self.utilheader_default_params.insert(0, "View Header")

      if protocol_yml["pipeline"][ite][0] == "UTIL_Merge":
        self.utilmerge_default_params = protocol_yml["pipeline"][ite]
        self.utilmerge_default_params.insert(0, "Merge Images")

      if protocol_yml["pipeline"][ite][0] == "QC_Report":
        self.qcreport_default_params = protocol_yml["pipeline"][ite]
        self.qcreport_default_params.insert(0, "QC Report")

    self.slicecheck_params = self.slicecheck_default_params.copy()
    self.interlacecheck_params = self.interlacecheck_default_params.copy()
    self.baselineaverage_params = self.baselineaverage_default_params.copy()
    self.susceptibility_params = self.susceptibility_default_params.copy()
    self.eddymotion_params = self.eddymotion_default_params.copy()
    self.brainmask_params = self.brainmask_default_params.copy()
    self.dtiestimate_params = self.dtiestimate_default_params.copy()
    self.exclude_params = self.exclude_default_params.copy()
    self.utilheader_params = self.utilheader_default_params.copy()
    self.utilmerge_params = self.utilmerge_default_params.copy()
    self.qcreport_params = self.qcreport_default_params.copy()


  def SetDicProtocol(self, new_dic_protocol):
    self.dic_protocol = new_dic_protocol.copy()

  def SetModuleId(self, identifier):
    self.module_id = identifier

  def SetLoadingProtocol(self, new_value):
    self.loading_protocol = new_value

  def SettingDefaultProtocol(self, module_name):
    if module_name ==  "SLICE_Check":
      self.slicecheck_params = self.slicecheck_default_params
    if module_name == "INTERLACE_Check":
      self.interlacecheck_params = self.interlacecheck_default_params
    if module_name == "BASELINE_Average":
      self.baselineaverage_params = self.baselineaverage_default_params
    if module_name == "SUSCEPTIBILITY_Correct":
      self.susceptibility_params = self.susceptibility_default_params
    if module_name == "EDDYMOTION_Correct":
      self.eddymotion_params = self.eddymotion_default_params
    if module_name == "BRAIN_Masking":
      self.brainmask_params = self.brainmask_default_params
    if module_name == "DTI_Estimate":
      self.dtiestimate_params = self.dtiestimate_default_params
    if module_name == "MANUAL_Exclude":
      self.exclude_params = self.exclude_default_params
    if module_name == "UTIL_Header":
      self.utilheader_params = self.utilheader_default_params
    if module_name == "UTIL_Merge":
      self.utilmerge_params = self.utilmerge_default_params
    if module_name == "QC_Report":
      self.qcreport_params = self.qcreport_default_params

  def RemoveModuleFromDicProtocol(self, item_to_remove_key):
    self.dic_protocol.pop(item_to_remove_key)
    




import numpy as np 
from dipy.align.imaffine import (transform_centers_of_mass,
                                 AffineMap,
                                 MutualInformationMetric,
                                 AffineRegistration)
import dipy.align 
from dipy.align.transforms import (TranslationTransform2D,
                                   RigidTransform2D,
                                   AffineTransform2D,
                                   TranslationTransform3D,
                                   RigidTransform3D,
                                   AffineTransform3D)
import dtiplayground.dmri.preprocessing as prep
import os 

logger=prep.logger.write


@prep.measure_time
def interlace_compute(image_obj):
    image_obj.images=image_obj.images.astype(float)
    # affine=np.transpose(np.append(image_obj.information['space_directions'],np.expand_dims(image_obj.information['space_origin'],0),axis=0))
    # affine=np.append(affine,np.array([[0,0,0,1]]),axis=0)
    affine=image_obj.getAffineMatrixForNifti()
    x,y,z,g = image_obj.images.shape
    evens=[x for x in range(0,z) if x % 2 == 0]
    odds=[x for x in range(0,z) if x % 2 ==1]
    if z % 2 == 1:
        evens.pop()

    output=[]
    for gidx in range(0,g):

        #### interlaing a volume
        static=image_obj.images[:,:,evens,gidx]
        moving=image_obj.images[:,:,odds,gidx]

        #### Correlation and Motion detection
        corr=ncc(moving,static)
        #logger("\rGradient {}/{}, Corr: {:.4f} , registering for motion detection ...".format(gidx,g,corr))
        transformed, out_affine=rigid_3d(static,moving,affine,affine ,
                                                      nbins=32,
                                                      level_iters=[10000,1000,100],
                                                      sigmas=[3.0,1.0,0.0],
                                                      factors=[4,2,1],sampling_prop=0.1)
        affine_info=decompose_affine_matrix(out_affine)
        max_norm=np.max(np.abs(affine_info["translations"]))
        max_angle_in_deg=np.max(np.rad2deg(np.abs(affine_info["angles"])))
        logger("Gradient {}/{}, Corr: {:.4f}, Max translation : {:.4f} , Max angle : {:.4f} degree".format(gidx,g-1,corr,max_norm,max_angle_in_deg))
        output.append({"gradient_index": gidx, 
                       "original_gradient_index": image_obj.getGradients()[gidx]['original_index'], 
                       "affine_matrix":out_affine.tolist(),
                       "correlation": float(corr),
                       "motions": affine_info})

    return output

@prep.measure_time
def interlace_check(image_obj, computation_result,
                    correlationDeviationBaseline=2.5,
                    correlationDeviationGradient=3.0,
                    correlationThresholdBaseline=0.95,
                    correlationThresholdGradient=0.7702,
                    rotationThreshold=0.5,
                    translationThreshold=1.5):
    result_size=len(computation_result)
    correlations=[x['correlation'] for x in computation_result]
    corr_avg=np.mean(correlations)
    corr_std=np.std(correlations)
    gradients=image_obj.getGradients()
    quadfit=quadratic_fit_generator(image_obj.getBValueBounds(),
                                    [min(correlationDeviationBaseline,correlationDeviationGradient),max(correlationDeviationBaseline,correlationDeviationGradient)]
                                    )
    excluding_gradients_indexes=[]
    interlacing_results=[]
    for idx,g in enumerate(computation_result):
        #logger("Checking Gradient {}/{} ".format(idx,result_size-1),prep.Color.PROCESS)
        result=g
        translation=np.max(np.abs(g['motions']['translations']))
        rotation_in_degree=np.max(np.rad2deg(np.abs(g['motions']['angles'])))
        correlation=g['correlation']
        corr_z=(correlation-corr_avg)/corr_std
        isB0=image_obj.isGradientBaseline(g['gradient_index'])
        bval=gradients[idx]['b_value']

        check_rotation= rotation_in_degree <= rotationThreshold
        check_translation= translation <= translationThreshold
        check_correlation= True
        quad_fitted_z_threshold=quadfit(bval)

        if isB0:
            check_correlation= not(corr_z < -quad_fitted_z_threshold)
        else:
            check_correlation= not(corr_z < -quad_fitted_z_threshold)

        excluded=not(check_rotation and check_translation and check_correlation)
        if not excluded:
            color=prep.Color.OK
        else:
            color=prep.Color.WARNING
            excluding_gradients_indexes.append(idx)
        logger("[Grad.idx:{0:03d}, Org.idx:{1:03d}]\tBaseline: {2}, Max.Trans: {3:.2f} , Max.Rotation: {4:.2f}deg, Corr: {5:.4f}, Corr-Z: {6:4f}"
            .format(g['gradient_index'],
                    g['original_gradient_index'],
                    isB0,
                    translation,
                    rotation_in_degree,
                    correlation,
                    corr_z
                    ),color)

        logger("\t\t\t\tZ-Threshold: {:.4f}, Check.Trans: {}, Check.Rotation: {}, Check.Corr: {}"
            .format(quad_fitted_z_threshold,
                    check_translation,
                    check_rotation,
                    check_correlation
                    ),color)

        g['max_translation']=float(translation )
        g['max_rotation_in_degree']=float(rotation_in_degree )
        g['check_translation']=bool(check_translation)
        g['check_rotation']=bool(check_rotation)
        g['check_correlation']=bool(check_correlation)
        g['excluded']=bool(excluded)
        g['corr_z_value']=float(corr_z)
        g['isB0']=bool(isB0)
        g['quad_fitted_z_threshold']=float(quad_fitted_z_threshold)
        g['b_value']=float(bval)
        interlacing_results.append(g)

    return excluding_gradients_indexes, interlacing_results
        

def _quad_fit(bval, domain=[0,1000], fimage=[3.0,3.5]): #returns std multiple between fimage
    if bval > domain[1] : 
        return fimage[1]
    elif bval < domain[0] : 
        return fimage[0]
    a= (fimage[1]-fimage[0])/((domain[1]-domain[0])**2)
    c= fimage[0]
    return a*(bval**2)+c

def quadratic_fit_generator(domain,fimage):
    def wrapper(bval):
        return _quad_fit(bval,domain,fimage)
    return wrapper

def ncc(x,y): #normalized cross correlation in image (ref: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3864968/ )
    ab=np.sum(x*y)
    a2=np.sum(x**2)
    b2=np.sum(y**2)
    if a2*b2==0.0: 
        return 1.0
    else:
        return ab/np.sqrt(a2*b2)

def affine_3d_to_2d(mat,indices=[0,1]):
    new_mat=mat.copy()
    to_remove=list(set(range(mat.shape[0]-1))-set(indices))
    new_mat=np.delete(new_mat,to_remove,axis=0)
    new_mat=np.transpose(np.delete(np.transpose(new_mat),to_remove,axis=0))
    return new_mat

def decompose_affine_matrix(mat4d): ## r
    scale, shear, angles, trans, persp =dipy.align.streamlinear.decompose_matrix(mat4d)
    angles_in_deg=list(map(lambda x : float(np.rad2deg(x)), angles))
    res={"scale":scale.tolist(),
         "shear":list(map(float,shear)),
         "angles":angles,
         "angles_in_degree":angles_in_deg,
         "translations":trans.tolist(),
         "perspective":persp.tolist()}
    return res

def measure_translation_from_affine_matrix(mat):
    translation=mat[:,2][:2]
    norm=np.sqrt(np.sum(translation*translation))
    return norm 

def measure_max_translation_from_affine_matrix(mat):
    translation=np.abs(mat[:,2][:2])
    norm=np.abs(np.max(translation))
    return norm 

@prep.measure_time
def rigid_3d(static,moving,
             affine_static,affine_moving,
             nbins=32,
             level_iters=[10000,1000,100],
             sigmas=[3.0,1.0,0.0],
             factors=[4,2,1],
             sampling_prop=None):
    ## Make affine map
    
    identity=np.eye(4)
    affine_map=AffineMap(identity,static.shape, affine_static,
                                  moving.shape, affine_static)

    ## Resample moving image from affine (but same in a volume)
    resampled=affine_map.transform(moving)

    ## center of mass transform
    c_of_mass = transform_centers_of_mass(static, affine_static, moving, affine_moving)
    transformed=c_of_mass.transform(moving)

    ## registration preparation
    metric= MutualInformationMetric(nbins,sampling_prop)
    affreg= AffineRegistration(metric=metric, 
                               level_iters=level_iters, 
                               sigmas=sigmas, 
                               factors=factors,
                               verbosity=0)

    ## Translation transform registration
    transform=TranslationTransform3D()
    params0=None
    starting_affine= c_of_mass.affine
    translation=affreg.optimize(static, 
                                moving, 
                                transform, 
                                params0, 
                                affine_static,affine_moving, 
                                starting_affine=starting_affine)
    transformed = translation.transform(moving)

    ## Ridid registration
    transform=RigidTransform3D()
    params0=None
    starting_affine=translation.affine
    rigid=affreg.optimize(static,moving,transform,params0,affine_static,affine_moving,starting_affine=starting_affine)
    transformed = rigid.transform(moving)

    return transformed, rigid.affine 



def rigid_2d(static,moving,
             affine_static,affine_moving,
             nbins=32,
             level_iters=[10000,1000,100],
             sigmas=[3.0,1.0,0.0],
             factors=[4,2,1],
             sampling_prop=None):
    ## Make affine map
    identity=np.eye(3)
    affine_map=AffineMap(identity,static.shape, affine_static,
                                  moving.shape, affine_moving)

    ## Resample moving image from affine (but same in a volume)
    resampled=affine_map.transform(moving)

    ## center of mass transform
    c_of_mass = transform_centers_of_mass(static, affine_static, moving, affine_moving)
    transformed=c_of_mass.transform(moving)

    ## registration preparation
    metric= MutualInformationMetric(nbins,sampling_prop)
    affreg= AffineRegistration(metric=metric, 
                               level_iters=level_iters, 
                               sigmas=sigmas, 
                               factors=factors,
                               verbosity=0)

    # Translation transform registration
    transform=TranslationTransform2D()
    params0=None
    starting_affine= c_of_mass.affine
    translation=affreg.optimize(static, 
                                moving, 
                                transform, 
                                params0, 
                                affine_static,affine_moving, 
                                starting_affine=starting_affine)
    transformed = translation.transform(moving)

    ## Ridid registration
    transform=RigidTransform2D()
    params0=None
    starting_affine=translation.affine
    rigid=affreg.optimize(static,moving,transform,params0,affine_static,affine_moving,starting_affine=starting_affine)
    transformed = rigid.transform(moving)

    return transformed, rigid.affine 



from dtiplayground.dmri.common.tools.base import ExternalToolWrapper

class FiberProcess(ExternalToolWrapper):
    def __init__(self,binary_path= None, **kwargs):
        super().__init__(binary_path, **kwargs)
        self.binary_path=None
        if binary_path is not None:
            self.binary_path=binary_path
        elif 'softwares' in kwargs:
            self.binary_path=kwargs['softwares']['fiberprocess']['path']




from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal as Signal


class ModulesDetailsCommunicate(QObject):

  update_params = Signal()
  send_params = Signal(list)
  call_update_user_preferences_file = Signal(dict)


  def __init__(self):
    super(ModulesDetailsCommunicate, self).__init__()

  def modif(self):
    self.update_params.emit()

  def SendParams(self, params):
    self.send_params.emit(params)

  def CallUpdateUserPreferencesFile(self, new_preferences):
    self.call_update_user_preferences_file.emit(new_preferences)

    



from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
from PyQt5.QtCore import * 
from pathlib import Path
import re
import os
from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal as Signal

from dtiplayground.ui.loadingbutton import LoadingButton
from dtiplayground.ui.quickview import QuickView

class ExecutionTabCommunicate(QObject):

  call_DeleteTempDirectory = Signal()
  call_UpdateNumberOfInputs = Signal(int)
  check_manual_exclude = Signal()
  call_compute = Signal(list)
  call_write_protocol_yml = Signal(str)
  check_unsaved_changes = Signal()
  call_save_protocol = Signal()

  def CallDeleteTempDirectory(self):
    self.call_DeleteTempDirectory.emit()

  def NumberOfInputs(self, multi_input):
    self.call_UpdateNumberOfInputs.emit(multi_input)

  def CheckManualExclude(self):
    self.check_manual_exclude.emit()

  def CallCompute(self, computation_details):
    self.call_compute.emit(computation_details)

  def CallWriteProtocolYML(self, temp_path):
    self.call_write_protocol_yml.emit(temp_path)

  def CheckUnsavedChangesAndStartCompute(self):
    self.check_unsaved_changes.emit()

  def CallSaveProtocol(self):
    self.call_save_protocol.emit()



class ExecutionTab(QWidget):
  communicate = ExecutionTabCommunicate()

  def __init__(self, protocol_template, args):
    QWidget.__init__(self)
    self.multi_input = 0
    self.args = args
    self.tab = QWidget() 
    self.communicate.call_UpdateNumberOfInputs.connect(self.UpdateNumberOfInputs)
    self.communicate.call_compute.connect(self.Compute)

    self.ExecutionTab(protocol_template)

  def ExecutionTab(self, protocol_template):

    layout = QGridLayout()
    self.input_layout = QGridLayout()
    
    self.input_line = QLineEdit()
    if hasattr(self, "input_filename"):
      self.input_line.setText(self.input_filename)
    self.input_line.setAlignment(Qt.AlignRight)
    self.input_line.editingFinished.connect(self.UpdateOutputFilenameBase)
    self.input_line_label = QLabel("Input image:")
    self.input_browse_button = QPushButton("Browse")
    self.input_browse_button.clicked.connect(self.BrowseButton0)
    self.input_layout.addWidget(self.input_line_label, 0, 0)
    self.input_layout.addWidget(self.input_line, 0, 1)
    self.input_layout.addWidget(self.input_browse_button, 0, 2)
    layout.addLayout(self.input_layout, 0, 0, 1, 6)

    self.output_image_basename = QLineEdit()
    self.output_image_basename.setAlignment(Qt.AlignRight)
    output_image_basename_label = QLabel("Output image basename")
    layout.addWidget(output_image_basename_label, 1, 0)
    layout.addWidget(self.output_image_basename, 1, 1, 1, 5)
    
    self.output_line = QLineEdit()
    self.output_line.setAlignment(Qt.AlignRight)
    self.output_line.setStatusTip(protocol_template["options"]["io"]["output_directory"]["description"])
    output_line_label = QLabel(protocol_template["options"]["io"]["output_directory"]["caption"])
    layout.addWidget(output_line_label, 2, 0)
    layout.addWidget(self.output_line, 2, 1, 1, 5)

    layout.addWidget(QLabel("Computation:"), 3, 0)
    self.cluster_computation = QRadioButton("Cluster / Slurm")
    layout.addWidget(self.cluster_computation, 3, 1)
    self.local_computation = QRadioButton("Local")
    self.local_computation.setChecked(True)
    layout.addWidget(self.local_computation, 3, 2)

    self.threads_line = QLineEdit()
    self.threads_line.setValidator(QIntValidator())
    self.threads_line.setAlignment(Qt.AlignRight)
    layout.addWidget(QLabel("Number of threads to use:"), 3, 4)
    layout.addWidget(self.threads_line, 3, 5)

    compute_button = LoadingButton("Compute")
    compute_button.clicked.connect(self.CheckInputFile)
    gif_path = Path(__file__).parent.joinpath('data/workinprogress.gif')
    compute_button.setGif(gif_path.__str__())
    layout.addWidget(compute_button, 4, 0, 1, 4)

    stop_compute_button = QPushButton("Stop Computation")
    layout.addWidget(stop_compute_button, 4, 4, 1, 2)

    self.output_display = QTextEdit(readOnly = True)
    self.output_display.setAcceptRichText(True)
    layout.addWidget(QLabel("Output:"), 5, 0)
    layout.addWidget(self.output_display, 6, 0, 1, 4)
    layout.setColumnStretch(2, 2)
    self.output_display.document().setMaximumBlockCount(1000000)

    self.error_display = QTextEdit(readOnly = True)
    layout.addWidget(QLabel("Error(s):"), 5, 4)
    layout.addWidget(self.error_display, 6, 4, 1, 2)
    layout.setColumnStretch(3, 0)

    self.process = QtCore.QProcess(self)
    self.process.readyReadStandardError.connect(self.DisplayError)
    self.process.readyReadStandardOutput.connect(self.DisplayOutput)    
    self.process.started.connect(lambda: compute_button.start())
    self.process.started.connect(lambda: compute_button.setEnabled(False))
    self.process.finished.connect(lambda: compute_button.setEnabled(True))
    self.process.finished.connect(lambda: compute_button.stop())
    self.process.finished.connect(self.communicate.CallDeleteTempDirectory)
    stop_compute_button.clicked.connect(self.process.kill)
    
    self.no_input_popup = QMessageBox()
    self.no_input_popup.setText("Missing input file.")
    self.no_input_popup.setWindowTitle("DMRIPrep message")

    self.unsaved_changes_popup = QMessageBox()
    self.unsaved_changes_popup.setWindowTitle("Unsaved changes")
    self.unsaved_changes_popup.setText("Unsaved modicfications have been made to the protocol. Do you want to save the modifications or ignore them?")
    self.unsaved_changes_popup.setStandardButtons(QMessageBox.Save | QMessageBox.Ignore)
    self.unsaved_changes_popup.setDefaultButton(QMessageBox.Save)
    self.unsaved_changes_popup.buttonClicked.connect(self.UnsavedPopupClicked)

    self.tab.setLayout(layout)


  def UpdateNumberOfInputs(self, multi_input):
    self.multi_input = multi_input

    # remove all widgets from QGridLayout input_layout 
    for i in reversed(range(self.input_layout.count())): 
      self.input_layout.itemAt(i).widget().setParent(None)
      
    # set new input_layout with 1 input file
    if self.multi_input == 0:
      self.input_line = QLineEdit()
      self.input_line.setAlignment(Qt.AlignRight)
      if hasattr(self, "input_filename"):
        self.input_line.setText(self.input_filename)
      self.input_line.editingFinished.connect(self.UpdateOutputFilenameBase)
      self.input_line_label = QLabel("Input image:")
      self.input_browse_button = QPushButton("Browse")
      self.input_browse_button.clicked.connect(self.BrowseButton0)
      self.input_layout.addWidget(self.input_line_label, 0, 0)
      self.input_layout.addWidget(self.input_line, 0, 1)
      self.input_layout.addWidget(self.input_browse_button, 0, 2)
      self.output_image_basename.setEnabled(False)
      
    # set new input_layout with 2 input files
    if self.multi_input == 1:  
      self.input_line_label1 = QLabel("Direction 1 input image:")
      self.input_layout.addWidget(self.input_line_label1, 0, 0)
      self.input_line1 = QLineEdit()
      self.input_line1.setAlignment(Qt.AlignRight)
      if hasattr(self, "input_filename1"):
        self.input_line1.setText(self.input_filename1)
      self.input_line1.editingFinished.connect(self.UpdateOutputFilenameBase)
      self.input_layout.addWidget(self.input_line1, 0, 1)
      self.input_browse_button1 = QPushButton("Browse")
      self.input_browse_button1.clicked.connect(self.BrowseButton1)
      self.input_layout.addWidget(self.input_browse_button1, 0, 2)
      self.input_line_label2 = QLabel("Direction 2 input image:")
      self.input_layout.addWidget(self.input_line_label2, 1, 0)
      self.input_line2 = QLineEdit()
      self.input_line2.setAlignment(Qt.AlignRight)
      if hasattr(self, "input_filename2"):
        self.input_line2.setText(self.input_filename2)
      self.input_layout.addWidget(self.input_line2, 1, 1)
      self.input_browse_button2 = QPushButton("Browse")
      self.input_browse_button2.clicked.connect(self.BrowseButton2)
      self.input_layout.addWidget(self.input_browse_button2, 1, 2)
      self.output_image_basename.setEnabled(True)

    # set new input_layout for UTIL_Merge
    if self.multi_input == 2:
      self.input_line_label = QLabel("Input images:")
      self.input_layout.addWidget(self.input_line_label, 0, 0)
      self.input_list_widget = QListWidget()
      self.input_list_widget.setFixedHeight(100)
      self.input_list_widget.setDragDropMode(QAbstractItemView.InternalMove)
      self.input_list_widget.setSelectionMode(QAbstractItemView.ExtendedSelection)
      self.input_list_widget.model().rowsMoved.connect(self.UpdateInputList)
      self.input_layout.addWidget(self.input_list_widget, 0, 1, 4, 1)
      if hasattr(self, "input_list"):
        self.AddInputToInputListWidget()
      self.merge_input_browse_button = QPushButton("Browse")
      self.input_layout.addWidget(self.merge_input_browse_button, 1, 0)
      self.merge_input_browse_button.clicked.connect(self.MergeBrowseButton)
      self.merge_input_remove_button = QPushButton("Remove")
      self.merge_input_remove_button.clicked.connect(self.RemoveInput)
      self.input_layout.addWidget(self.merge_input_remove_button, 2, 0)
      self.merge_input_clear_button = QPushButton("Clear")
      self.merge_input_clear_button.clicked.connect(self.ClearInput)
      self.input_layout.addWidget(self.merge_input_clear_button, 3, 0)
      self.output_image_basename.setEnabled(True)

    
  def UpdateOutputFilenameBase(self):
    if self.multi_input == 0: # single input
      input_path = self.input_filename
    elif self.multi_input == 1: # multi input (SUSCEPTIBILITY_Correction for now)
      input_path = self.input_filename1
    elif self.multi_input == 2: # UTIL_Merge
      if len(self.input_list) > 0: 
        input_path = self.input_list[0]
      else:
        self.output_image_basename.setText("")
        return
    image = input_path.split("/")[-1]
    basename = image.split(".")[0]
    self.output_image_basename.setText(basename)

  def CheckInputFile(self):
    if (self.multi_input == 0 and self.input_line.text() == "") or (self.multi_input == 1 and (self.input_line1.text() == "" or self.input_line2.text() == "")) or (self.multi_input == 2 and self.input_list_widget.count() < 2):
      self.no_input_popup.exec_()
    else:
      self.communicate.CheckManualExclude()

  def Compute(self, computation_details):
    
    manual_exclude = computation_details[0] #True if MANUAL_Exclude in protocol, False if not
    quickview = computation_details[1] #True if QuickView selected, False if not
    
    if manual_exclude and quickview:
      self.quickview_window = QuickView(self.input_filename)
      self.quickview_window.show()

    if manual_exclude and not quickview: #update dic_protocol, write temp protocol, start compute temp protocol
      if os.path.exists("temp_dmriprep_ui"):
        print("Directory 'temp_dmriprep_ui' already exists")   
      else:
        os.mkdir("temp_dmriprep_ui")
        print("Directory 'temp_dmriprep_ui' created")
      self.communicate.CallWriteProtocolYML("temp_dmriprep_ui/protocol.yml")
      self.StartComputation("temp_dmriprep_ui/protocol.yml")
    if not manual_exclude and not quickview: 
      self.communicate.CheckUnsavedChangesAndStartCompute()

  def StartComputation(self, protocol):
    arguments = ["run", "-i"] 
    if self.multi_input == 0:
      arguments.append(self.input_filename)
    elif self.multi_input == 1:
      arguments.append(self.input_filename1)
      arguments.append(self.input_filename2)
    elif self.multi_input == 2:
      arguments += self.input_list
    arguments.append("-p")
    arguments.append(protocol)
    arguments.append("-o")
    if self.multi_input == 0:
      inputpath = self.input_filename
    elif self.multi_input == 1:
      inputpath = self.input_filename1
    elif self.multi_input ==2:
      inputpath = self.input_list[0]
    if self.output_line.text() != '':
      if str(Path(inputpath).parent) not in self.output_line.text():
        outputpath = str(Path(inputpath).parent) + "/" + self.output_line.text()
      else:
        outputpath = self.output_line.text()
    else: #set default output directory
      suffix = "_dmriprep-dev_output"
      if ".nrrd" in inputpath:
        outputpath = inputpath.replace(".nrrd", suffix)
      if ".nii.gz" in inputpath:
        outputpath = inputpath.replace(".nii.gz", suffix)
      if ".nii" in inputpath:
        outputpath = inputpath.replace(".nii", suffix)
      #check that directory does not already exist
      counter = 0
      new_outputpath = outputpath
      while os.path.exists(new_outputpath):
        new_outputpath = outputpath + "_" + str(counter)
        counter += 1
      outputpath = new_outputpath
      self.output_line.setText(outputpath)

    arguments.append(outputpath)
    if self.output_image_basename.text != '':
      arguments.append("--output-file-base")
      arguments.append(self.output_image_basename.text())
    if self.threads_line.text() != '':
      arguments.append("--num-threads")
      arguments.append(self.threads_line.text())

    if self.local_computation.isChecked():
      self.process.start("dmriprep", arguments)
    if self.cluster_computation.isChecked():
      arguments[6] = arguments[6].split("/")[-1]
      arguments = ["longleaf.unc.edu", "'", "/proj/NIRAL/containers/dtiplaygroung/dmriprep_longleaf"] + arguments + ["'"]
      self.process.start("ssh", arguments)

  def UnsavedPopupClicked(self, i):
    if i.text() == "Save":
      self.communicate.CallSaveProtocol()

  def BrowseButton0(self):
    text = self.BrowseInputFile()
    self.input_line.setText(text)
    if text != "":
      self.input_filename = text
      self.UpdateOutputFilenameBase()

  def BrowseButton1(self):
    text = self.BrowseInputFile()
    self.input_line1.setText(text)
    if text != "":
      self.input_filename1 = text
      self.UpdateOutputFilenameBase()

  def BrowseButton2(self):
    text = self.BrowseInputFile()
    self.input_line2.setText(text)
    if text != "":
      self.input_filename2 = text

  def BrowseInputFile(self):
    if self.args.output_directory:
      self.output_line.setText(self.args.output_directory)
    else:
      self.output_line.setText("")
    file_filter = "Image file (*.nrrd *.nii *.nii.gz)"
    file_name = QFileDialog.getOpenFileName(
      parent = self,
      caption = "Select an input file",
      filter = file_filter
      )
    return file_name[0]

  def MergeBrowseButton(self):
    if self.args.output_directory:
      self.output_line.setText(self.args.output_directory)
    else:
      self.output_line.setText("")
    file_filter = "Image files (*.nrrd *.nii *.nii.gz)"
    file_names = QFileDialog.getOpenFileNames(
      parent = self,
      caption = "Select an input file",
      filter = file_filter
      )
    if len(file_names[0]) != 0:
      if not hasattr(self, "input_list"):
        self.input_list = []
      self.input_list += file_names[0]
      self.AddInputToInputListWidget()
      self.UpdateOutputFilenameBase()

  def RemoveInput(self):
    items = self.input_list_widget.selectedItems()
    for i in items:
      removed_item_index = self.input_list_widget.indexFromItem(i).row()
      removed_item = self.input_list_widget.takeItem(removed_item_index)
    self.UpdateInputList()

  def UpdateInputList(self):
    self.input_list = []
    for i in range(self.input_list_widget.count()):
      self.input_list.append(self.input_list_widget.item(i).text())
    self.UpdateOutputFilenameBase()

  def ClearInput(self):
    if hasattr(self, "input_list"):    
      for ite in range(self.input_list_widget.count()):
        removed_item = self.input_list_widget.takeItem(0)
      self.UpdateInputList()

  def AddInputToInputListWidget(self):
    for ite in range(self.input_list_widget.count()):
      removed_item = self.input_list_widget.takeItem(0)
    for input_filename in self.input_list:
      self.input_list_widget.addItem(input_filename)

  def DisplayOutput(self):
    stdout = self.process.readAllStandardOutput()
    stdout = stdout.data().decode("utf8")

    # colors
    stdout = re.sub(r"\\[95m", "<p style='color:lightpurple;'>", stdout)
    stdout = re.sub(r"\\[35m", "<p style='color:purple;'>", stdout)
    stdout = re.sub(r"\\[96m", "<p style='color:lightcyan;'>", stdout)
    stdout = re.sub(r"\\[36m", "<p style='color:cyan;'>", stdout)
    stdout = re.sub(r"\\[94m", "<p style='color:lightblue;'>", stdout)
    stdout = re.sub(r"\\[34m", "<p style='olor:blue;'>", stdout)
    stdout = re.sub(r"\\[92m", "<p style='color:lightgreen;'>", stdout)
    stdout = re.sub(r"\\[32m", "<p style='color:green;'>", stdout)
    stdout = re.sub(r"\\[93m", "<p style='color:lightyellow;'>", stdout)
    stdout = re.sub(r"\\[33m", "<p style='color:yellow;'>", stdout)
    stdout = re.sub(r"\\[91m", "<p style='color:lightred;'>", stdout)
    stdout = re.sub(r"\\[31m", "<p style='color:red;'>", stdout)    
    stdout = re.sub(r"\\[90m", "<p style='color:gray;'>", stdout)
    stdout = re.sub(r"\\[30m", "<p style='color:black;'>", stdout)
    # style
    stdout = re.sub(r"\\[1m", "<p style='font-weight: 900;'>", stdout)
    stdout = re.sub(r"\\[4m", "<p style='text-decoration: underline;'>", stdout)
    stdout = re.sub(r"\\[0m", "</span>", stdout)
    stdout = re.sub(r"\n", "</p><p>", stdout)
    stdout = re.sub(r"\t", "<span STYLE='padding:0 0 0 20px;'>", stdout)
    self.output_display.insertHtml(stdout)
    self.output_display.verticalScrollBar().setValue(self.output_display.verticalScrollBar().maximum())

  def DisplayError(self):
    stderr = self.process.readAllStandardError()
    stderr = bytes(stderr).decode("utf8")
    # colors
    stdout = re.sub(r"\\[95m", "<p style='color:lightpurple;'>", stderr)
    stdout = re.sub(r"\\[35m", "<p style='color:purple;'>", stderr)
    stdout = re.sub(r"\\[96m", "<p style='color:lightcyan;'>", stderr)
    stdout = re.sub(r"\\[36m", "<p style='color:cyan;'>", stderr)
    stdout = re.sub(r"\\[94m", "<p style='color:lightblue;'>", stderr)
    stdout = re.sub(r"\\[34m", "<p style='olor:blue;'>", stderr)
    stdout = re.sub(r"\\[92m", "<p style='color:lightgreen;'>", stderr)
    stdout = re.sub(r"\\[32m", "<p style='color:green;'>", stderr)
    stdout = re.sub(r"\\[93m", "<p style='color:lightyellow;'>", stderr)
    stdout = re.sub(r"\\[33m", "<p style='color:yellow;'>", stderr)
    stdout = re.sub(r"\\[91m", "<p style='color:lightred;'>", stderr)
    stdout = re.sub(r"\\[31m", "<p style='color:red;'>", stderr)    
    stdout = re.sub(r"\\[90m", "<p style='color:gray;'>", stderr)
    stdout = re.sub(r"\\[30m", "<p style='color:black;'>", stderr)
    # style
    stdout = re.sub(r"\\[1m", "<p style='font-weight: 900;'>", stderr)
    stdout = re.sub(r"\\[4m", "<p style='text-decoration: underline;'>", stderr)
    stdout = re.sub(r"\\[0m", "</p>", stderr)

    self.error_display.setHtml(stderr)


#!python

import shutil
import os 
import importlib
import traceback,time,copy,yaml,sys,uuid
from pathlib import Path
import argparse,yaml
from argparse import RawTextHelpFormatter
sys.path.append(Path(__file__).parent.__str__())


import dtiplayground.dmri.common 
import dtiplayground.dmri.common.study as study
import dtiplayground.dmri.common.study.loaders as study_loaders
from dtiplayground.config import INFO as info
logger=dtiplayground.dmri.common.logger.write 


### decorators

def log_off(func):
    def wrapper(*args,**kwargs):
        dtiplayground.dmri.common.logger.setVerbosity(False)
        res=func(*args,**kwargs)
        dtiplayground.dmri.common.logger.setVerbosity(True)
        return res 
    return wrapper


### command functions

def command_import(args):
    ## reparametrization
    options={
        "study_dirs" : args.study_dirs,
        "execution_id":args.execution_id,
        "output" : args.output
    }
    if options['output'] is None:
        options['output']='output.yml'

    ## load config file and run pipeline
    res = study_loaders.load_bids(options['study_dirs'])
    with open(options['output'],'w') as f:
        yaml.dump(res,f)
    return res

### Arguments 

def get_args():
    current_dir=Path(__file__).parent
    # info=yaml.safe_load(open(current_dir.parent.joinpath('dtiplayground/info.json'),'r'))
    version=info['dtiplayground']['version']
    logger("VERSION : {}".format(str(version)))
    config_dir=Path(os.environ.get('HOME')).joinpath('.niral-dti')
    uid, ts = dtiplayground.dmri.common.get_uuid(), dtiplayground.dmri.common.get_timestamp()

    ### Argument parsers

    parser=argparse.ArgumentParser(prog="dtiplayground",
                                   formatter_class=RawTextHelpFormatter,
                                   description="DTIPlayground is a integration software for the overall DTI processes such as dmriprep, dtiatlasbuilder and so on.",
                                   epilog="Written by SK Park (sangkyoon_park@med.unc.edu) , Neuro Image Research and Analysis Laboratories, University of North Carolina @ Chapel Hill , United States, 2021")
    #parser.add_argument('command',help='command',type=str)
    subparsers=parser.add_subparsers(help="Commands")
    

    ## run command
    parser_import=subparsers.add_parser('import',help='Run Test')
    parser_import.add_argument('-s','--study-dirs',help='Study Directory',type=str,nargs='+')
    parser_import.add_argument('-o','--output',help="Output Filename",type=str,required=False)
    parser_import.set_defaults(func=command_import)

    ## log related
    parser.add_argument('--config-dir',help='Configuration directory',default=str(config_dir))
    parser.add_argument('--log',help='log file',default=str(config_dir.joinpath('log.txt')))
    parser.add_argument('--system-log-dir',help='System log directory',default='/BAND/USERS/skp78-dti/system-logs',type=str)
    parser.add_argument('--execution-id',help='execution id',default=uid,type=str)
    parser.add_argument('--no-log-timestamp',help='Remove timestamp in the log', default=False, action="store_true")
    parser.add_argument('--no-verbosity',help='Do not show any logs in the terminal', default=False, action="store_true")
    

    ## if no parameter is furnished, exit with printing help
    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args=parser.parse_args()

    ## system log
    sys_log_dir=Path(args.system_log_dir)
    sys_log_dir.mkdir(parents=True,exist_ok=True)
    env=os.environ
    
    sys_logfilename='dtiplayground'+env['USER']+"_"+ts+"_"+args.execution_id+".txt"
    sys_logfile=sys_log_dir.joinpath(sys_logfilename)
    dtiplayground.dmri.common.logger.addLogfile(sys_logfile.__str__(),mode='w')
    logger("Execution ID : {}".format(args.execution_id))
    logger("Execution Command : "+" ".join(sys.argv))

    return args 

args=get_args()

if __name__=='__main__':
    try:
        dtiplayground.dmri.common.logger.setTimestamp(True)
        result=args.func(args)
        exit(0)
    except Exception as e:
        dtiplayground.dmri.common.logger.setVerbosity(True)
        msg=traceback.format_exc()
        logger(msg,dtiplayground.dmri.common.Color.ERROR)
        exit(-1)
    finally:
        pass





from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import *
from PyQt5.QtCore import * 
from PIL import Image, ImageEnhance
from PIL.ImageQt import ImageQt
from functools import partial
import SimpleITK as sitk
import numpy
import os

from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal as Signal


class QuickViewCommunicate(QObject):

  execute_exclude_gradients = Signal(list)

  def __init__(self):
    super(QuickViewCommunicate, self).__init__()

  def Execution(self, list_of_gradients):
    self.execute_exclude_gradients.emit(list_of_gradients)


class QuickView(QMainWindow):
  signal_quickview = QuickViewCommunicate()

  def __init__(self, image_name):
    QMainWindow.__init__(self)

    self.setWindowTitle("QuickView")
    widget = QWidget()
    self.setCentralWidget(widget)

    quickview_layout = QHBoxLayout()
    widget.setLayout(quickview_layout)

    images_groupbox = QGroupBox()
    self.images_layout = QGridLayout()
    images_groupbox.setLayout(self.images_layout)

    options_groupbox = QGroupBox()
    self.options_layout = QVBoxLayout()
    options_groupbox.setLayout(self.options_layout)
    
    gradients_groupbox = QGroupBox()
    self.gradients_layout = QVBoxLayout()
    gradients_groupbox.setLayout(self.gradients_layout)

    self.menu_layout = QVBoxLayout()
    self.menu_layout.addWidget(options_groupbox)
    self.menu_layout.addWidget(gradients_groupbox)  
    
    self.image_name = image_name

    self.LoadImage()
    self.Menu()
    if ".nrrd" in self.image_name:
      self.CreateButtonsNRRD()
    else:
      self.CreateButtonsNIFTI()
    
    scroll = QScrollArea()
    scroll.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
    scroll.setWidgetResizable(True)
    scroll.setWidget(images_groupbox)

    quickview_layout.addWidget(scroll, 80)
    quickview_layout.addLayout(self.menu_layout, 20)

    self.show()
    
  def Menu(self):   
    self.columns = 9 #default
    max_slice = self.size[0] #default
    self.zslice = self.size[0]//2 #default
    self.gradients_layout.addWidget(QLabel("Selected gradients:"))
    self.gradientsToExclude_list = []
    self.gradientsToExclude_label = QLabel(str(self.gradientsToExclude_list))
    self.gradientsToExclude_label.setWordWrap(True)
    self.gradients_layout.addWidget(self.gradientsToExclude_label)
    self.gradients_layout.addStretch(1)
    
    compute_button = QPushButton("Exclude Gradients")
    compute_button.clicked.connect(self.ComputeExcludeGradients)
    self.gradients_layout.addWidget(compute_button)
    
    self.options_layout.addWidget(QLabel("View:"))
    self.sagittal = QRadioButton("Sagittal View")
    self.sagittal.clicked.connect(self.ChangeView)
    self.sagittal.setChecked(True) #default
    self.options_layout.addWidget(self.sagittal)
    self.coronal = QRadioButton("Coronal View")
    self.coronal.clicked.connect(self.ChangeView)
    self.options_layout.addWidget(self.coronal)
    self.axial = QRadioButton("Axial View")
    self.axial.clicked.connect(self.ChangeView)
    self.options_layout.addWidget(self.axial)
    self.options_layout.addStretch(1)

    self.options_layout.addWidget(QLabel("Slice:"))
    self.slice_selector_slidebar = QSlider(Qt.Horizontal)
    self.slice_selector_slidebar.setSingleStep(1)
    self.slice_selector_slidebar.setMinimum(0)
    self.slice_selector_slidebar.setMaximum(max_slice) #default
    self.slice_selector_slidebar.setValue(self.zslice) #default
    self.slice_selector_slidebar.valueChanged.connect(lambda: self.slice_selector_spinbox.setValue(self.slice_selector_slidebar.value()))
    self.slice_selector_slidebar.sliderReleased.connect(self.ChangeSlice)
    self.options_layout.addWidget(self.slice_selector_slidebar)
    self.slice_selector_spinbox = QSpinBox()
    self.slice_selector_spinbox.setAlignment(Qt.AlignRight)
    self.slice_selector_spinbox.setSingleStep(1)
    self.slice_selector_spinbox.setMinimum(0)
    self.slice_selector_spinbox.setMaximum(max_slice) #default
    self.slice_selector_spinbox.setValue(self.zslice) #default
    self.slice_selector_spinbox.valueChanged.connect(lambda: self.slice_selector_slidebar.setValue(self.slice_selector_spinbox.value()))
    self.slice_selector_spinbox.valueChanged.connect(self.ChangeSlice)
    self.options_layout.addWidget(self.slice_selector_spinbox)
    self.options_layout.addStretch(1)

    self.options_layout.addWidget(QLabel("Number of columns:"))
    self.images_per_row = QSpinBox()
    self.images_per_row.setMinimum(1)
    if ".nrrd" in self.image_name:
      self.images_per_row.setMaximum(self.number_gradients)
    else:
      self.images_per_row.setMaximum(self.size[3])
    self.images_per_row.setAlignment(Qt.AlignRight)
    self.images_per_row.setValue(self.columns)
    self.images_per_row.valueChanged.connect(self.ChangeColumns)
    self.options_layout.addWidget(self.images_per_row)
    self.options_layout.addStretch(1)

    self.zoom_factor = 1
    self.options_layout.addWidget(QLabel("Zoom factor:"))
    self.zoom = QDoubleSpinBox()
    self.zoom.setValue(self.zoom_factor)
    self.zoom.valueChanged.connect(self.ChangeZoom)
    self.options_layout.addWidget(self.zoom)
    self.options_layout.addStretch(1)

    self.brightness_factor = 1
    self.options_layout.addWidget(QLabel("Brightness:"))
    self.brightness_slidebar = QSlider(Qt.Horizontal)
    self.brightness_slidebar.setMinimum(0)
    self.brightness_slidebar.setMaximum(100) 
    self.brightness_slidebar.setValue(self.brightness_factor*50) 
    self.brightness_slidebar.valueChanged.connect(self.ChangeBrightness)
    self.options_layout.addWidget(self.brightness_slidebar)
    self.options_layout.addStretch(1)

    self.contrast_factor = 1
    self.options_layout.addWidget(QLabel("Contrast:"))
    self.contrast_slidebar = QSlider(Qt.Horizontal)
    self.contrast_slidebar.setMinimum(0)
    self.contrast_slidebar.setMaximum(100) 
    self.contrast_slidebar.setValue(self.contrast_factor*50) 
    self.contrast_slidebar.valueChanged.connect(self.ChangeContrast)
    self.options_layout.addWidget(self.contrast_slidebar)
    self.options_layout.addStretch(1)

    apply_button = QPushButton("&Apply")
    apply_button.setShortcut('Enter')
    apply_button.clicked.connect(self.RemoveImages)
    self.options_layout.addWidget(apply_button)

  def LoadImage(self):
    print("image", self.image_name)
    self.input_image = sitk.ReadImage(self.image_name)
    self.size = list(self.input_image.GetSize())
    self.spacing = list(self.input_image.GetSpacing())
    self.number_gradients = self.input_image.GetNumberOfComponentsPerPixel()

  def AxialViewNRRD(self):
    self.number_slices = self.size[2]
    index = [0, 0, self.zslice]
    spacing_matrix = QTransform()
    spacing_matrix.scale(self.spacing[0], self.spacing[1])
    slice_extractor = sitk.ExtractImageFilter()
    slice_extractor.SetSize([self.size[0], self.size[1], 0])
    slice_extractor.SetIndex(index)
    extracted_slice = slice_extractor.Execute(self.input_image)
    return extracted_slice, spacing_matrix

  def SagittalViewNRRD(self):
    self.number_slices = self.size[0]
    index = [self.zslice, 0, 0, ]
    spacing_matrix = QTransform()
    spacing_matrix.scale(self.spacing[1], self.spacing[2])
    slice_extractor = sitk.ExtractImageFilter()
    slice_extractor.SetSize([0, self.size[1], self.size[2]])
    slice_extractor.SetIndex(index)
    extracted_slice = slice_extractor.Execute(self.input_image)
    return extracted_slice, spacing_matrix

  def CoronalViewNRRD(self):
    self.number_slices = self.size[1]
    index = [0, self.zslice, 0]
    spacing_matrix = QTransform()
    spacing_matrix.scale(self.spacing[0], self.spacing[2])
    slice_extractor = sitk.ExtractImageFilter()    
    slice_extractor.SetSize([self.size[0], 0, self.size[2]])
    slice_extractor.SetIndex(index)
    extracted_slice = slice_extractor.Execute(self.input_image)
    return extracted_slice, spacing_matrix
    
  def CreateButtonsNRRD(self):

    self.dic = {}
    
    if self.axial.isChecked():
      extracted_slice, spacing_matrix = self.AxialViewNRRD()
    if self.sagittal.isChecked():
      extracted_slice, spacing_matrix = self.SagittalViewNRRD()
    if self.coronal.isChecked():
      extracted_slice, spacing_matrix = self.CoronalViewNRRD()

    for iter_gradients in range(self.number_gradients):
      self.button = QPushButton()
      self.dic[str(iter_gradients)] = self.button
      
      gradient_extractor = sitk.VectorIndexSelectionCastImageFilter()
      gradient_extractor.SetIndex(iter_gradients)
      gradient = gradient_extractor.Execute(extracted_slice)

      gradient_array = sitk.GetArrayFromImage(gradient)

      # min max scaling for brightness
      gradient_array_normalized = (gradient_array - numpy.min(gradient_array)) * round(255 / numpy.max(gradient_array), 3)
      if self.sagittal.isChecked() or self.coronal.isChecked():
        gradient_array_normalized = numpy.flipud(gradient_array_normalized) #if SAGITTAL or CORONAL
      gradient_image = Image.fromarray(gradient_array_normalized)
      gradient_image = gradient_image.convert("L")
      gradient_image2 = gradient_image.resize((int(self.zoom_factor*gradient_image.width), int(self.zoom_factor*gradient_image.height)))
      # brightness and contrast
      brightness_enhancer = ImageEnhance.Brightness(gradient_image2)
      gradient_image3 = brightness_enhancer.enhance(self.brightness_factor)
      contrast_enhancer = ImageEnhance.Contrast(gradient_image3)
      gradient_image4 = contrast_enhancer.enhance(self.contrast_factor)

      gradient_qimage = ImageQt(gradient_image4)
      gradient_qimage.transformed(spacing_matrix)
      gradient_qpixmap = QtGui.QPixmap.fromImage(gradient_qimage)  

      self.button.setIcon(QIcon(gradient_qpixmap))
      self.button.setIconSize(gradient_qpixmap.size())
      self.images_layout.addWidget(QLabel("DWI " + str(iter_gradients)), 2*(iter_gradients//self.columns), iter_gradients%self.columns, QtCore.Qt.AlignCenter)
      self.images_layout.addWidget(self.button, 2*(iter_gradients//self.columns)+1, iter_gradients%self.columns)
      self.button.clicked.connect(partial(self.ButtonClicked, dwi=iter_gradients)) 
      if iter_gradients in self.gradientsToExclude_list:
        self.button.setStyleSheet("background-color : red")
      else:
        self.button.setStyleSheet("background-color : black") 
    
  def SagittalViewNIFTI(self, iter_gradients):
    self.number_slices = self.size[0]
    index = [self.zslice, 0, 0, iter_gradients]
    spacing_matrix = QTransform()
    spacing_matrix.scale(self.spacing[1], self.spacing[2])
    slice_extractor = sitk.ExtractImageFilter()
    slice_extractor.SetSize([0, self.size[1], self.size[2], 0])
    slice_extractor.SetIndex(index)
    gradient = slice_extractor.Execute(self.input_image)
    return gradient, spacing_matrix

  def CoronalViewNIFTI(self, iter_gradients):
    self.number_slices = self.size[1]
    index = [0, self.zslice, 0, iter_gradients]
    spacing_matrix = QTransform()
    spacing_matrix.scale(self.spacing[0], self.spacing[2])
    slice_extractor = sitk.ExtractImageFilter()    
    slice_extractor.SetSize([self.size[0], 0, self.size[2], 0])
    slice_extractor.SetIndex(index)
    gradient = slice_extractor.Execute(self.input_image)
    return gradient, spacing_matrix

  def AxialViewNIFTI(self, iter_gradients):
    self.number_slices = self.size[2]
    index = [0, 0, self.zslice, iter_gradients]
    spacing_matrix = QTransform()
    spacing_matrix.scale(self.spacing[0], self.spacing[1])
    slice_extractor = sitk.ExtractImageFilter()
    slice_extractor.SetSize([self.size[0], self.size[1], 0, 0])
    slice_extractor.SetIndex(index)
    gradient = slice_extractor.Execute(self.input_image)
    return gradient, spacing_matrix

  def CreateButtonsNIFTI(self):
    
    self.dic = {}
    for iter_gradients in range(self.size[3]):
      self.button = QPushButton()
      self.dic[str(iter_gradients)] = self.button
      
      if self.sagittal.isChecked():
        gradient, spacing_matrix = self.SagittalViewNIFTI(iter_gradients)
      if self.coronal.isChecked():
        gradient, spacing_matrix = self.CoronalViewNIFTI(iter_gradients)
      if self.axial.isChecked():
        gradient, spacing_matrix = self.AxialViewNIFTI(iter_gradients)

      gradient_array = sitk.GetArrayFromImage(gradient)

      # min max scaling for brightness
      gradient_array_normalized = (gradient_array - numpy.min(gradient_array)) * round(255 / numpy.max(gradient_array), 3)
      if self.sagittal.isChecked() or self.coronal.isChecked():
        gradient_array_normalized = numpy.flipud(gradient_array_normalized) #if SAGITTAL or CORONAL
      gradient_image = Image.fromarray(gradient_array_normalized)
      gradient_image = gradient_image.convert("L")
      gradient_image2 = gradient_image.resize((int(self.zoom_factor*gradient_image.width), int(self.zoom_factor*gradient_image.height)))
      
      # brightness and contrast
      brightness_enhancer = ImageEnhance.Brightness(gradient_image2)
      gradient_image3 = brightness_enhancer.enhance(self.brightness_factor)
      contrast_enhancer = ImageEnhance.Contrast(gradient_image3)
      gradient_image4 = contrast_enhancer.enhance(self.contrast_factor)

      gradient_qimage = ImageQt(gradient_image4)
      gradient_qimage.transformed(spacing_matrix)
      gradient_qpixmap = QtGui.QPixmap.fromImage(gradient_qimage)  

      self.button.setIcon(QIcon(gradient_qpixmap))
      self.button.setIconSize(gradient_qpixmap.size())
      self.images_layout.addWidget(QLabel("DWI " + str(iter_gradients)), 2*(iter_gradients//self.columns), iter_gradients%self.columns, QtCore.Qt.AlignCenter)
      self.images_layout.addWidget(self.button, 2*(iter_gradients//self.columns)+1, iter_gradients%self.columns)
      self.button.clicked.connect(partial(self.ButtonClicked, dwi=iter_gradients)) 
      if iter_gradients in self.gradientsToExclude_list:
        self.button.setStyleSheet("background-color : red")
      else:
        self.button.setStyleSheet("background-color : black") 

  def ButtonClicked(self, dwi):
    if dwi not in self.gradientsToExclude_list:
      self.gradientsToExclude_list.append(dwi)
      self.dic[str(dwi)].setStyleSheet("background-color : red")
    else:
      self.gradientsToExclude_list.remove(dwi)
      self.dic[str(dwi)].setStyleSheet("background-color : black")

    self.gradientsToExclude_label.setText(str(self.gradientsToExclude_list))

  def ChangeView(self):
    self.RemoveImages()
    if self.axial.isChecked():
      self.slice_selector_spinbox.setMaximum(self.size[2])
      self.slice_selector_slidebar.setMaximum(self.size[2])
      if self.zslice >= self.size[2]:
        self.zslice = self.size[2]-1
        self.slice_selector_spinbox.setValue(self.zslice)
    if self.sagittal.isChecked():
      self.slice_selector_spinbox.setMaximum(self.size[0])
      self.slice_selector_slidebar.setMaximum(self.size[0])
      if self.zslice >= self.size[0]:
        self.zslice = self.size[0]-1
        self.slice_selector_spinbox.setValue(self.zslice)
    if self.coronal.isChecked():
      self.slice_selector_spinbox.setMaximum(self.size[1])
      self.slice_selector_slidebar.setMaximum(self.size[1])
      if self.zslice >= self.size[1]:
        self.zslice = self.size[1]-1
        self.slice_selector_spinbox.setValue(self.zslice)
    if ".nrrd" in self.image_name:
      self.CreateButtonsNRRD()
    else:
      self.CreateButtonsNIFTI()

  def ChangeSlice(self):
    self.RemoveImages()
    self.zslice = self.slice_selector_spinbox.value()
    if ".nrrd" in self.image_name:
      self.CreateButtonsNRRD()
    else:
      self.CreateButtonsNIFTI()
  
  def ChangeColumns(self):
    self.RemoveImages()
    self.columns = self.images_per_row.value()
    if ".nrrd" in self.image_name:
      self.CreateButtonsNRRD()
    else:
      self.CreateButtonsNIFTI()

  def ChangeZoom(self):
    self.RemoveImages()
    self.zoom_factor = self.zoom.value()
    if ".nrrd" in self.image_name:
      self.CreateButtonsNRRD()
    else:
      self.CreateButtonsNIFTI()

  def ChangeBrightness(self):
    self.RemoveImages()
    self.brightness_factor = round(self.brightness_slidebar.value()/50, 2)
    if ".nrrd" in self.image_name:
      self.CreateButtonsNRRD()
    else:
      self.CreateButtonsNIFTI()

  def ChangeContrast(self):
    self.RemoveImages()
    self.contrast_factor = round(self.contrast_slidebar.value()/50, 2)
    if ".nrrd" in self.image_name:
      self.CreateButtonsNRRD()
    else:
      self.CreateButtonsNIFTI()

  def RemoveImages(self):

    # remove everything from viewing window
    for i in reversed(range(self.images_layout.count())):
      widget_to_remove = self.images_layout.itemAt(i).widget()
      self.images_layout.removeWidget(widget_to_remove) #remove from layout list
      widget_to_remove.setParent(None) #remove from GUI

    self.zslice = self.slice_selector_spinbox.value()

    if self.axial.isChecked():
      self.slice_selector_spinbox.setMaximum(self.size[2])
      self.slice_selector_slidebar.setMaximum(self.size[2])
      if self.zslice >= self.size[2]:
        self.zslice = self.size[2]-1
        self.slice_selector_spinbox.setValue(self.zslice)
    if self.sagittal.isChecked():
      self.slice_selector_spinbox.setMaximum(self.size[0])
      self.slice_selector_slidebar.setMaximum(self.size[0])
      if self.zslice >= self.size[0]:
        self.zslice = self.size[0]-1
        self.slice_selector_spinbox.setValue(self.zslice)
    if self.coronal.isChecked():
      self.slice_selector_spinbox.setMaximum(self.size[1])
      self.slice_selector_slidebar.setMaximum(self.size[1])
      if self.zslice >= self.size[1]:
        self.zslice = self.size[1]-1
        self.slice_selector_spinbox.setValue(self.zslice)

    self.columns = self.images_per_row.value()

    self.zoom_factor = self.zoom.value()
    
    if ".nrrd" in self.image_name:
      self.CreateButtonsNRRD()
    else:
      self.CreateButtonsNIFTI()

  def ComputeExcludeGradients(self):
    self.signal_quickview.Execution(self.gradientsToExclude_list)
    
    self.close()


    
  
